# encoding: utf-8
from __future__ import print_function

from PIL import ImageOps, ImageChops
from collections import defaultdict, OrderedDict
from copy import copy
from enum import unique
from functools import wraps

try:
    from functools import reduce
except ImportError:
    pass

from instakit.abc import Enum, Fork, NOOp, Sequence, MutableSequence
from instakit.utils.gcr import BasicGCR
from instakit.utils.mode import Mode
from instakit.utils.misc import string_types, tuplize
from instakit.processors.adjust import AutoContrast

__all__ = ('Pipe', 'Pipeline',
           'BandFork', 'ChannelFork',
           'CMYKInk', 'RGBInk',
           'OverprintFork', 'ChannelOverprinter')

__dir__ = lambda: list(__all__)

if not hasattr(__builtins__, 'cmp'):
    def cmp(a, b):
        return (a > b) - (a < b)

class Pipe(Sequence):
    
    """ A static linear pipeline of processors to be applied en masse.
        Derived from a `pilkit` class:
            `pilkit.processors.base.ProcessorPipeline`
    """
    __slots__ = ('tuple',)
    
    @classmethod
    def base_type(cls):
        return tuple
    
    @wraps(tuple.__init__)
    def __init__(self, *args):
        self.tuple = tuplize(*args)
    
    def iterate(self):
        return iter(self.tuple)
    
    @wraps(tuple.__len__)
    def __len__(self):
        return len(self.tuple)
    
    @wraps(tuple.__contains__)
    def __contains__(self, value):
        return value in self.tuple
    
    @wraps(tuple.__getitem__)
    def __getitem__(self, idx):
        return self.tuple[idx]
    
    @wraps(tuple.index)
    def index(self, value):
        return self.tuple.index(value)
    
    def last(self):
        if not bool(self):
            raise IndexError("pipe is empty")
        return self.tuple[-1]
    
    def process(self, image):
        for processor in self.iterate():
            image = processor.process(image)
        return image
    
    def __eq__(self, other):
        if not isinstance(other, (type(self), type(self).base_type())):
            return NotImplemented
        return 

class Pipeline(MutableSequence):
    
    """ A mutable linear pipeline of processors to be applied en masse.
        Derived from a `pilkit` class:
            `pilkit.processors.base.ProcessorPipeline`
    """
    __slots__ = ('list',)
    
    @classmethod
    def base_type(cls):
        return list
    
    @wraps(list.__init__)
    def __init__(self, *args):
        base_type = type(self).base_type()
        if len(args) == 0:
            self.list = base_type()
        if len(args) == 1:
            target = args[0]
            if type(target) is type(self):
                self.list = copy(target.list)
            elif type(target) is base_type:
                self.list = copy(target)
            elif type(target) in (tuple, set, frozenset):
                self.list = base_type([*target])
            elif type(target) in (dict, defaultdict, OrderedDict):
                self.list = base_type([*sorted(target).values()])
            elif hasattr(target, 'iterate'):
                self.list = base_type([*target.iterate()])
            elif hasattr(target, '__iter__'):
                self.list = base_type([*target])
            else:
                self.list = base_type([target])
        else:
            self.list = base_type([*args])
    
    def iterate(self):
        return iter(self.list)
    
    @wraps(list.__len__)
    def __len__(self):
        return len(self.list)
    
    @wraps(list.__contains__)
    def __contains__(self, value):
        return value in self.list
    
    @wraps(list.__getitem__)
    def __getitem__(self, idx):
        return self.list[idx]
    
    @wraps(list.__setitem__)
    def __setitem__(self, idx, value):
        if value in (None, NOOp):
            value = NOOp()
        self.list[idx] = value
    
    @wraps(list.__delitem__)
    def __delitem__(self, idx):
        del self.list[idx]
    
    @wraps(list.index)
    def index(self, value):
        return self.list.index(value)
    
    @wraps(list.append)
    def append(self, value):
        self.list.append(value)
    
    @wraps(list.extend)
    def extend(self, iterable):
        self.list.extend(iterable)
    
    def pop(self, idx=-1):
        """ Remove and return item at `idx` (default last).
            Raises IndexError if list is empty or `idx` is out of range.
            See list.pop(…) for details.
        """
        self.list.pop(idx)
    
    def last(self):
        if not bool(self):
            raise IndexError("pipe is empty")
        return self.list[-1]
    
    def process(self, image):
        for processor in self.iterate():
            image = processor.process(image)
        return image

class BandFork(Fork):
    
    """ BandFork is a processor container -- a processor that applies other
        processors. BandFork acts selectively on the individual bands of
        input image data, either:
        - applying a band-specific processor instance, or
        - applying a default processor factory successively across all bands.
        
        BandFork’s interface is closely aligned with Python’s mutable-mapping
        API‡ -- with which most programmers are no doubt quite familiar:
        
        • Ex. 1: apply Atkinson dithering to each of an RGB images’ bands:
        >>> from instakit.utils.pipeline import BandFork
        >>> from instakit.processors.halftone import Atkinson
        >>> BandFork(Atkinson).process(my_image)
        
        • Ex. 2: apply Atkinson dithering to only the green band:
        >>> from instakit.utils.pipeline import BandFork
        >>> from instakit.processors.halftone import Atkinson
        >>> bfork = BandFork(None)
        >>> bfork['G'] = Atkinson()
        >>> bfork.process(my_image)
        
        BandFork inherits from `instakit.abc.Fork`, which itself is not just
        an Instakit Processor. The Fork ABC implements the required methods
        of an Instakit Processor Container†, through which it furnishes an
        interface to individual bands -- also generally known as channels,
        per the language of the relevant Photoshop UI elements -- of image
        data. 
        
        † q.v. the `instakit.abc` module source code supra.
        ‡ q.v. the `collections.abc` module, and the `MutableMapping`
                    abstract base class within, supra.
    """
    mode_t = Mode.RGB
    
    def __init__(self, processor_factory, *args, **kwargs):
        """ Initialize a BandFork instance, using the given callable value
            for `processor_factory` and any band-appropriate keyword-arguments,
            e.g. `(R=MyProcessor, G=MyOtherProcessor, B=None)`
        """
        # Reset `self.mode` if a new mode was specified:
        if 'mode' in kwargs:
            self.mode = kwargs.pop('mode')
        
        # Call `super(…)`, passing `processor_factory`:
        super(BandFork, self).__init__(processor_factory, *args, **kwargs)
    
    @property
    def mode(self):
        return self.mode_t
    
    @mode.setter
    def mode(self, value):
        if type(value) in string_types:
            value = Mode.for_string(value)
        if type(value) is Mode:
            if value is not self.mode_t:
                self.set_mode_t(value)
        else:
            raise TypeError("invalid mode type: %s (%s)" % (type(value), value))
    
    def set_mode_t(self, value):
        self.mode_t = value # DOUBLE SHADOW!!
    
    @property
    def band_labels(self):
        return self.mode_t.bands
    
    def iterate(self):
        for band_label in self.band_labels:
            yield self[band_label]
    
    def split(self, image):
        return self.mode_t.process(image).split()
    
    def compose(self, *bands):
        return self.mode_t.merge(*bands)
    
    def process(self, image):
        processed = []
        for processor, band in zip(self.iterate(),
                                   self.split(image)):
            processed.append(processor.process(band))
        return self.compose(*processed)

ChannelFork = BandFork

ink_values = (
    (255, 255, 255),    # White
    (0,   250, 250),    # Cyan
    (250, 0,   250),    # Magenta
    (250, 250, 0),      # Yellow
    (0,   0,   0),      # Key (blacK)
    (255, 0,   0),      # Red
    (0,   255, 0),      # Green
    (0,   0,   255),    # Blue
)

class Ink(Enum):
    __slots__ = tuple()
    
    def rgb(self):
        return ink_values[self.value]
    
    def process(self, image):
        InkType = type(self)
        return ImageOps.colorize(Mode.L.process(image),
                                 InkType(0).rgb(),
                                 InkType(self.value).rgb())

@unique
class CMYKInk(Ink):
    __slots__ = tuple()
    
    WHITE = 0
    CYAN = 1
    MAGENTA = 2
    YELLOW = 3
    KEY = 4
    
    @classmethod
    def CMYK(cls):
        return (cls.CYAN, cls.MAGENTA, cls.YELLOW, cls.KEY)
    
    @classmethod
    def CMY(cls):
        return (cls.CYAN, cls.MAGENTA, cls.YELLOW)

@unique
class RGBInk(Ink):
    __slots__ = tuple()
    
    WHITE = 0
    RED = 5
    GREEN = 6
    BLUE = 7
    KEY = 4
    
    @classmethod
    def RGB(cls):
        return (cls.RED, cls.GREEN, cls.BLUE)
    
    @classmethod
    def BGR(cls):
        return (cls.BLUE, cls.GREEN, cls.RED)

class OverprintFork(BandFork):
    
    """ A BandFork subclass that rebuilds its output image using multiply-mode
        to simulate CMYK overprinting effects.
        
        N.B. While this Fork-based processor operates strictly in CMYK mode,
        the composite image it eventually returns will be in RGB mode. This is
        because the CMYK channels are each individually converted to colorized
        representations in order to simulate monotone ink preparations; the
        final compositing operation, in which these colorized channel separation
        images are combined with multiply-mode, is also computed using the RGB
        color model -- q.v. the CMYKInk enum processor supra. and the related
        PIL/Pillow module function `ImageOps.colorize(…)` supra.
    """
    __slots__ = ('contrast', 'basicgcr')
    
    mode_t = Mode.CMYK
    inks = CMYKInk.CMYK()
    
    def __init__(self, processor_factory, gcr=20, *args, **kwargs):
        """ Initialize an OverprintFork instance with the given callable value
            for `processor_factory` and any band-appropriate keyword-arguments,
            e.g. `(C=MyProcessor, M=MyOtherProcessor, Y=MyProcessor, K=None)`
        """
        # Store BasicGCR and AutoContrast processors:
        self.contrast = AutoContrast()
        self.basicgcr = BasicGCR(percentage=gcr)
        
        # Call `super(…)`, passing `processor_factory`:
        super(OverprintFork, self).__init__(processor_factory, *args, **kwargs)
        
        # Make each band-processor a Pipeline() ending in
        # the channel-appropriate CMYKInk enum processor:
        self.apply_CMYK_inks()
    
    def apply_CMYK_inks(self):
        """ This method ensures that each bands’ processor is set up
            as a Pipe() or Pipeline() ending in a CMYKInk corresponding
            to the band in question. Calling it multiple times *should*
            be idempotent (but don’t quote me on that)
        """
        for band_label, ink in zip(self.band_labels,
                              type(self).inks):
            processor = self[band_label]
            if processor is None:
                self[band_label] = Pipe(ink)
            elif hasattr(processor, 'append'):
                if processor[-1] is not ink:
                    processor.append(ink)
                    self[band_label] = processor
            elif hasattr(processor, 'last'):
                if processor.last() is not ink:
                    self[band_label] = Pipe(*processor.iterate(), ink)
            else:
                self[band_label] = Pipe(processor, ink)
    
    def set_mode_t(self, value):
        """ Raise an exception if an attempt is made to set the mode to anything
            other than CMYK
        """
        if value is not Mode.CMYK:
            raise AttributeError(
                "OverprintFork only works in %s mode" % Mode.CMYK.to_string())
    
    def update(self, iterable=None, **kwargs):
        """ OverprintFork.update(…) re-applies CMYK ink processors to the
            updated processing dataflow
        """
        super(OverprintFork, self).update(iterable, **kwargs)
        self.apply_CMYK_inks()
    
    def split(self, image):
        """ OverprintFork.split(image) uses imagekit.utils.gcr.BasicGCR(…) to perform
            gray-component replacement in CMYK-mode images; for more information,
            see the imagekit.utils.gcr module
        """
        return self.basicgcr.process(image).split()
    
    def compose(self, *bands):
        """ OverprintFork.compose(…) uses PIL.ImageChops.multiply() to create
            the final composite image output
        """
        return reduce(ImageChops.multiply, bands)

class Grid(Fork):
    pass

class Sequence(Fork):
    pass

ChannelOverprinter = OverprintFork

if __name__ == '__main__':
    from pprint import pprint
    from instakit.utils.static import asset
    from instakit.processors.halftone import Atkinson
    
    image_paths = list(map(
        lambda image_file: asset.path('img', image_file),
            asset.listfiles('img')))
    image_inputs = list(map(
        lambda image_path: Mode.RGB.open(image_path),
            image_paths))
    
    for image_input in image_inputs[:2]:
        OverprintFork(Atkinson).process(image_input).show()
        
        print('Creating OverprintFork and BandFork with Atkinson ditherer...')
        overatkins = OverprintFork(Atkinson)
        forkatkins = BandFork(Atkinson)
        
        print('Processing image with BandForked Atkinson in default (RGB) mode...')
        forkatkins.process(image_input).show()
        forkatkins.mode = 'CMYK'
        print('Processing image with BandForked Atkinson in CMYK mode...')
        forkatkins.process(image_input).show()
        forkatkins.mode = 'RGB'
        print('Processing image with BandForked Atkinson in RGB mode...')
        forkatkins.process(image_input).show()
        
        overatkins.mode = 'CMYK'
        print('Processing image with OverprintFork-ized Atkinson in CMYK mode...')
        overatkins.process(image_input).show()
        
        print('Attempting to reset OverprintFork to RGB mode...')
        import traceback, sys
        try:
            overatkins.mode = 'RGB'
            overatkins.process(image_input).show()
        except:
            print(">>>>>>>>>>>>>>>>>>>>> TRACEBACK <<<<<<<<<<<<<<<<<<<<<")
            traceback.print_exc(file=sys.stdout)
            print("<<<<<<<<<<<<<<<<<<<<< KCABECART >>>>>>>>>>>>>>>>>>>>>")
            print('')
    
    bandfork = BandFork(None)
    pprint(bandfork)
    
    print(image_paths)
    



#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False

from cython.operator cimport address
from instakit.utils.ext cimport funcs

cpdef double[:] hsluv_to_rgb(double[:] hsl_triple) nogil:
    cdef double[:] rgb_triple = hsl_triple
    funcs.hsluv2rgb(        hsl_triple[0],          hsl_triple[1],          hsl_triple[2],
                    address(rgb_triple[0]), address(rgb_triple[1]), address(rgb_triple[2]))
    return rgb_triple

cpdef double[:] rgb_to_hsluv(double[:] rgb_triple) nogil:
    cdef double[:] hsl_triple = rgb_triple
    funcs.rgb2hsluv(        rgb_triple[0],          rgb_triple[1],          rgb_triple[2],
                    address(hsl_triple[0]), address(hsl_triple[1]), address(hsl_triple[2]))
    return hsl_triple

cpdef double[:] hpluv_to_rgb(double[:] hpl_triple) nogil:
    cdef double[:] rgb_triple = hpl_triple
    funcs.hpluv2rgb(        hpl_triple[0],          hpl_triple[1],          hpl_triple[2],
                    address(rgb_triple[0]), address(rgb_triple[1]), address(rgb_triple[2]))
    return rgb_triple

cpdef double[:] rgb_to_hpluv(double[:] rgb_triple) nogil:
    cdef double[:] hpl_triple = rgb_triple
    funcs.rgb2hpluv(        rgb_triple[0],          rgb_triple[1],          rgb_triple[2],
                    address(hpl_triple[0]), address(hpl_triple[1]), address(hpl_triple[2]))
    return hpl_triple



#!/usr/bin/env python
# encoding: utf-8
"""
utils/colortype.py

Created by FI$H 2000 on 2012-08-23.
Copyright (c) 2012 Objects In Space And Time, LLC. All rights reserved.
"""

import numpy
from collections import namedtuple, defaultdict
from instakit.utils.mode import split_abbreviations

color_types = defaultdict(dict)

# hash_RGB = lambda rgb: (rgb[0]*256)**2 + (rgb[1]*256) + rgb[2]

def ColorType(name, *args, **kwargs):
    global color_types
    dtype = numpy.dtype(kwargs.pop('dtype', numpy.uint8))
    if name not in color_types[dtype.name]:
        channels = split_abbreviations(name)
        
        class Color(namedtuple(name, channels)):
            
            def __repr__(self):
                return "%s(dtype=%s, %s)" % (
                    name, self.__class__.dtype.name,
                    ', '.join(['%s=%s' % (i[0], i[1]) \
                        for i in self._asdict().items()]))
            
            def __hex__(self):
                return '0x' + "%x" * len(self) % self
            
            def __int__(self):
                return int(self.__hex__(), 16)
            
            def __long__(self):
                return numpy.long(self.__hex__(), 16)
            
            def __hash__(self):
                return self.__long__()
            
            def __eq__(self, other):
                if not len(other) == len(self):
                    return False
                return all([self[i] == other[i] for i in range(len(self))])
            
            def __str__(self):
                return str(repr(self))
            
            def composite(self):
                return numpy.dtype([
                    (k, self.__class__.dtype) for k, v in self._asdict().items()])
            
        Color.__name__ = "%s<%s>" % (name, dtype.name)
        Color.dtype = dtype
        color_types[dtype.name][name] = Color
    return color_types[dtype.name][name]

if __name__ == '__main__':
    
    assert split_abbreviations('RGB') == ('R', 'G', 'B')
    assert split_abbreviations('CMYK') == ('C', 'M', 'Y', 'K')
    assert split_abbreviations('YCbCr') == ('Y', 'Cb', 'Cr')
    assert split_abbreviations('sRGB') == ('R', 'G', 'B')
    assert split_abbreviations('XYZ') == ('X', 'Y', 'Z')
    


#!/usr/bin/env python
# encoding: utf-8
"""
ndarrays.py

Created by FI$H 2000 on 2013-08-23.
Copyright © 2012-2019 Objects In Space And Time, LLC. All rights reserved.

The `bytescale`[0], `fromimage`[1], and `toimage`[2] functions have been
adapted from the versions published in the now-historic `scipy.misc.pilutils`
module; the last official release of which looks to have been in SciPy 1.1.0:

* [0] https://git.io/fhIoX
* [1] https://git.io/fhIo1
* [2] https://git.io/fhIoD

"""
from __future__ import division, print_function

import numpy
from instakit.utils.mode import Mode
from instakit.abc import NDProcessorBase

uint8_t = numpy.uint8
uint32_t = numpy.uint32
float32_t = numpy.float32

def bytescale(data, cmin=None, cmax=None,
                    high=255,  low=0):
    """
    Byte-scales a `numpy.ndarray` of nd-image data.
    
    “Byte scaling” means 1) casting the input image to the ``uint8_t`` dtype, and
                         2) scaling the range to ``(low, high)`` (default 0-255).
    
    If the input image data is already of dtype ``uint8_t``, no scaling is done.
    
    Parameters
    ----------
    data : `numpy.ndarray`
        PIL image data array.
    cmin : scalar, optional
        Bias scaling of small values. Default is ``data.min()``.
    cmax : scalar, optional
        Bias scaling of large values. Default is ``data.max()``.
    high : scalar, optional
        Scale max value to `high`.  Default is 255.
    low : scalar, optional
        Scale min value to `low`.  Default is 0.
    
    Returns
    -------
    array : `numpy.ndarray` of dtype ``uint8_t``
        The byte-scaled array.
    """
    if data.dtype == uint8_t:
        return data
    
    if high > 255:
        raise ValueError("`high` should be less than or equal to 255.")
    if low < 0:
        raise ValueError("`low` should be greater than or equal to 0.")
    if high < low:
        raise ValueError("`high` should be greater than or equal to `low`.")
    
    if cmin is None:
        cmin = data.min()
    
    if cmax is None:
        cmax = data.max()
    
    cscale = cmax - cmin
    
    if cscale < 0:
        raise ValueError("`cmax` should be larger than `cmin`.")
    elif cscale == 0:
        cscale = 1
    
    scale = float(high - low) / cscale
    bytedata = (data - cmin) * scale + low
    return (bytedata.clip(low, high) + 0.5).astype(uint8_t)

def fromimage(image, flatten=False,
                        mode=None,
                       dtype=None):
    """
    Return the data from an input PIL image as a `numpy.ndarray`.
    
    Parameters
    ----------
    im : PIL image
        Input image.
    flatten : bool, optional
        If true, convert the output to greyscale. Default is False.
    mode : str / Mode, optional
        Mode to convert image to, e.g. ``'RGB'``. See the Notes of the
        `imread` docstring for more details.
    dtype : str / ``numpy.dtype``, optional
        Numpy dtype to which to cast the output image array data,
        e.g. ``'float64'`` or ``'uint16'``. 
    
    Returns
    -------
    fromimage : ndarray (rank 2..3)
        The individual color channels of the input image are stored in the
        third dimension, such that greyscale (`L`) images are MxN (rank-2),
        `RGB` images are MxNx3 (rank-3), and `RGBA` images are MxNx4 (rank-3).
    """
    from PIL import Image
    
    if not Image.isImageType(image):
        raise TypeError("Input is not a PIL image (got %s)" % repr(image))
    
    if mode is not None:
        if not Mode.is_mode(mode):
            mode = Mode.for_string(mode)
        image = mode.process(image)
    elif Mode.of(image) is Mode.P:
        # Mode 'P' means there is an indexed "palette".  If we leave the mode
        # as 'P', then when we do `a = numpy.array(im)` below, `a` will be a 2D
        # containing the indices into the palette, and not a 3D array
        # containing the RGB or RGBA values.
        if 'transparency' in image.info:
            image = Mode.RGBA.process(image)
        else:
            image = Mode.RGB.process(image)
    
    if flatten:
        image = Mode.F.process(image)
    elif Mode.of(image) is Mode.MONO:
        # Workaround for crash in PIL. When im is 1-bit, the call numpy.array(im)
        # can cause a seg. fault, or generate garbage. See
        # https://github.com/scipy/scipy/issues/2138 and
        # https://github.com/python-pillow/Pillow/issues/350.
        # This converts im from a 1-bit image to an 8-bit image.
        image = Mode.L.process(image)
    
    out = numpy.array(image)
    
    if dtype is not None:
        return out.astype(
              numpy.dtype(dtype))
    
    return out

_errstr = "Mode unknown or incompatible with input array shape"

def toimage(array,  high=255,    low=0,
                    cmin=None,  cmax=None,
                     pal=None,
                    mode=None,
            channel_axis=None):
    """
    Takes an input `numpy.ndarray` and returns a PIL image.
            
    The mode of the image returned depends on 1) the array shape, and 
                                              2) the `pal` and `mode` keywords.
    
    For 2D arrays, if `pal` is a valid (N, 3) rank-2, ``uint8_t`` bytearray --
    populated with an `RGB` LUT of values from 0 to 255, ``mode='P'`` (256-color
    single-channel palette mode) will be used; otherwise ``mode='L'`` (256-level
    single-channel grayscale mode) will be employed -- unless a “mode” argument
    is given, as either 'F' or 'I'; in which case conversion to either a float
    or an integer rank-3 array will be made.
    
    .. warning::
    
        This function calls `bytescale` under the hood, to rescale the image
        pixel values across the full (0, 255) ``uint8_t`` range if ``mode``
        is one of either: ``None, 'L', 'P', 'l'``.
        
        It will also cast rank-2 image data to ``uint32_t`` when ``mode=None``
        (which is the default).
    
    Notes
    -----
    For 3D arrays, the `channel_axis` argument tells which dimension of the
    array holds the channel data. If one of the dimensions is 3, the mode
    is 'RGB' by default, or 'YCbCr' if selected.
    
    The input `numpy.ndarray` must be either rank-2 or rank-3.
    """
    from pprint import pformat
    
    data = numpy.asarray(array)
    if numpy.iscomplexobj(data):
        raise ValueError("Cannot convert arrays of complex values")
    
    shape = list(data.shape)
    valid = len(shape) == 2 or ((len(shape) == 3) and
                                ((3 in shape) or (4 in shape)))
    if not valid:
        raise ValueError("input array lacks a suitable shape for any mode")
    
    if mode is not None:
        if not Mode.is_mode(mode):
            mode = Mode.for_string(mode)
    
    if len(shape) == 2:
        shape = (shape[1], shape[0])  # columns show up first
        
        if mode is Mode.F:
            return mode.frombytes(shape, data.astype(float32_t).tostring())
        
        if mode in [ None, Mode.L, Mode.P ]:
            bytedata = bytescale(data, high=high,
                                       low=low,
                                       cmin=cmin,
                                       cmax=cmax)
            image = Mode.L.frombytes(shape, bytedata.tostring())
            
            if pal is not None:
                image.putpalette(numpy.asarray(pal,
                                 dtype=uint8_t).tostring()) # Becomes mode='P' automatically
            elif mode is Mode.P:  # default grayscale
                pal = (numpy.arange(0, 256, 1, dtype=uint8_t)[:, numpy.newaxis] *
                       numpy.ones((3,),        dtype=uint8_t)[numpy.newaxis, :])
                image.putpalette(numpy.asarray(pal,
                                 dtype=uint8_t).tostring())
            
            return image
        
        if mode is Mode.MONO:  # high input gives threshold for 1
            bytedata = (data > high)
            return mode.frombytes(shape, bytedata.tostring())
        
        if cmin is None:
            cmin = numpy.amin(numpy.ravel(data))
        
        if cmax is None:
            cmax = numpy.amax(numpy.ravel(data))
        
        data = (data * 1.0 - cmin) * (high - low) / (cmax - cmin) + low
        
        if mode is Mode.I:
            image = mode.frombytes(shape, data.astype(uint32_t).tostring())
        else:
            raise ValueError(_errstr)
        
        return image
    
    # if here then 3D array with a 3 or a 4 in the shape length.
    # Check for 3 in datacube shape --- 'RGB' or 'YCbCr'
    if channel_axis is None:
        if (3 in shape):
            ca = numpy.flatnonzero(numpy.asarray(shape) == 3)[0]
        else:
            ca = numpy.flatnonzero(numpy.asarray(shape) == 4)
            if len(ca):
                ca = ca[0]
            else:
                raise ValueError(
                    "Could not find a channel dimension (shape = %s)" % pformat(shape))
    else:
        ca = channel_axis
    
    numch = shape[ca]
    if numch not in [3, 4]:
        raise ValueError("Channel dimension invalid (#channels = %s)" % numch)
    
    bytedata = bytescale(data, high=high,
                               low=low,
                               cmin=cmin,
                               cmax=cmax)
    
    if ca == 2:
        strdata = bytedata.tostring()
        shape = (shape[1], shape[0])
    elif ca == 1:
        strdata = numpy.transpose(bytedata, (0, 2, 1)).tostring()
        shape = (shape[2], shape[0])
    elif ca == 0:
        strdata = numpy.transpose(bytedata, (1, 2, 0)).tostring()
        shape = (shape[2], shape[1])
    
    if mode is None:
        if numch == 3:
            mode = Mode.RGB
        else:
            mode = Mode.RGBA
    
    if mode not in [ Mode.RGB, Mode.RGBA, Mode.YCbCr, Mode.CMYK ]:
        raise ValueError(_errstr)
    
    if mode in [ Mode.RGB, Mode.YCbCr ]:
        if numch != 3:
            raise ValueError("Invalid shape for mode “%s”: %s" % (
                              mode, pformat(shape)))
    if mode in [ Mode.RGBA, Mode.CMYK ]:
        if numch != 4:
            raise ValueError("Invalid shape for mode “%s”: %s" % (
                              mode, pformat(shape)))
    
    # Here we know both `strdata` and `mode` are correct:
    image = mode.frombytes(shape, strdata)
    return image


class NDProcessor(NDProcessorBase):
    
    """ An image processor ancestor class that represents PIL image
        data in a `numpy.ndarray`. Subclasses can override the
        `process_nd(…)` method to receive, transform, and return
        the image data using NumPy, SciPy, and the like.
    """
    __slots__ = tuple()
    
    def process(self, image):
        """ NDProcessor.process(…) converts its PIL image operand
            to a `numpy.ndarray`, hands it off to the delegate
            method NDProcessor.process_nd(…), and converts whatever
            that method call returns back to a PIL image before
            finally returning it.
        """
        return toimage(self.process_nd(fromimage(image)))
    
    @staticmethod
    def compand(ndimage):
        """ The NDProcessor.compand(…) static method scales a
            `numpy.ndarray` with floating-point values from 0.0»1.0
            to unsigned 8-bit integer values from 0»255.
        """
        return uint8_t(float32_t(ndimage) * 255.0)
    
    @staticmethod
    def uncompand(ndimage):
        """ The NDProcessor.uncompand(…) static method scales a
            `numpy.ndarray` with unsigned 8-bit integer values
            from 0»255 to floating-point values from 0.0»1.0.
        """
        return float32_t(ndimage) / 255.0


def test():
    """ Tests for bytescale(¬) adapted from `scipy.misc.pilutil` doctests,
        q.v. https://git.io/fhkHI supra.
    """
    from instakit.utils.static import asset
    
    image_paths = list(map(
        lambda image_file: asset.path('img', image_file),
            asset.listfiles('img')))
    image_inputs = list(map(
        lambda image_path: Mode.RGB.open(image_path),
            image_paths))
    
    # print()
    print("«TESTING: BYTESCALE UTILITY FUNCTION»")
    
    image = numpy.array((91.06794177,   3.39058326,  84.4221549,
                         73.88003259,  80.91433048,   4.88878881,
                         51.53875334,  34.45808177,  27.5873488)).reshape((3, 3))
    assert numpy.all(
           bytescale(image) == numpy.array((255,   0, 236,
                                            205, 225,   4,
                                            140,  90,  70),
                                           dtype=uint8_t).reshape((3, 3)))
    assert numpy.all(
           bytescale(image,
                     high=200,
                     low=100) == numpy.array((200, 100, 192,
                                              180, 188, 102,
                                              155, 135, 128),
                                             dtype=uint8_t).reshape((3, 3)))
    assert numpy.all(
           bytescale(image,
                     cmin=0,
                     cmax=255) == numpy.array((91,  3, 84,
                                               74, 81,  5,
                                               52, 34, 28),
                                              dtype=uint8_t).reshape((3, 3)))
    print("«SUCCESS»")
    print()
    
    # print()
    print("«TESTING: FROMIMAGE»")
    
    image_arrays = list(map(
        lambda image_input: fromimage(image_input),
            image_inputs))
    
    for idx, image_array in enumerate(image_arrays):
        assert image_array.dtype == uint8_t
        assert image_array.shape[0] == image_inputs[idx].size[1]
        assert image_array.shape[1] == image_inputs[idx].size[0]
    
    print("«SUCCESS»")
    print()

if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
"""
                                                                                
                        d8888 888888b.    .d8888b.                              
      o                d88888 888  "88b  d88P  Y88b                    o        
     d8b              d88P888 888  .88P  888    888                   d8b       
    d888b            d88P 888 8888888K.  888        .d8888b          d888b      
"Y888888888P"       d88P  888 888  "Y88b 888        88K          "Y888888888P"  
  "Y88888P"        d88P   888 888    888 888    888 "Y8888b.       "Y88888P"    
  d88P"Y88b       d8888888888 888   d88P Y88b  d88P      X88       d88P"Y88b    
 dP"     "Yb     d88P     888 8888888P"   "Y8888P"   88888P'      dP"     "Yb   
                                                                                
                                                                                
Instakit’s Abstract Base Classes – née ABCs – for processors and data structures

"""
from __future__ import print_function

from pkgutil import extend_path
from abc import ABC, abstractmethod as abstract
from collections import defaultdict
from enum import Enum as EnumBase
from itertools import chain

if '__path__' in locals():
    __path__ = extend_path(__path__, __name__)

__all__ = ('is_in_class', 'subclasshook',
           'Processor', 'Enum', 'NOOp',
           'Container', 'MutableContainer',
           'Fork',
           'ThresholdMatrixProcessor',
           'NDProcessorBase')

__dir__ = lambda: list(__all__)

def is_in_class(attr, cls):
    """ Test whether or not a class has a named attribute,
        regardless of whether the class uses `__slots__` or
        an internal `__dict__`.
    """
    if hasattr(cls, '__dict__'):
        return attr in cls.__dict__
    elif hasattr(cls, '__slots__'):
        return attr in cls.__slots__
    return False

def subclasshook(cls, subclass):
    """ A subclass hook function for both Processor and Enum """
    if subclass in (Processor, Enum):
        if any(is_in_class('process', ancestor) for ancestor in subclass.__mro__):
            return True
    return NotImplemented

def slots_for(cls):
    """ Get the summation of the __slots__ tuples for a class and its ancestors """
    # q.v. https://stackoverflow.com/a/6720815/298171
    return tuple(chain.from_iterable(
                 getattr(ancestor, '__slots__', tuple()) \
                           for ancestor in cls.__mro__))

def compare_via_attrs(self, other):
    """ Compare two processors:
        1) Return NotImplemented if the types do not exactly match.
        2) For processors using __dict__ mappings for attributes,
           compare them directly.
        3) For processors using __slots__ for attributes,
           iterate through all ancestor slot names using “slots_for(…)”
           and return False if any compare inequal between self and other --
           ultimately returning True.
        4) If the slots/dict situation differs between the two instances,
           raise a TypeError.
    """
    if type(self) is not type(other):
        return NotImplemented
    if hasattr(self, '__dict__') and hasattr(other, '__dict__'):
        return self.__dict__ == other.__dict__
    elif hasattr(self, '__slots__') and hasattr(other, '__slots__'):
        slots = slots_for(type(self))
        for slot in slots:
            if getattr(self, slot) != getattr(other, slot):
                return False
        return True
    raise TypeError("dict/slots mismatch")

class Processor(ABC):
    
    """ Base abstract processor class. """
    __slots__ = tuple()
    
    @abstract
    def process(self, image):
        """ Process an image instance, per the processor instance,
            returning the processed image data.
        """
        ...
    
    def __call__(self, image):
        return self.process(image)
    
    @classmethod
    def __subclasshook__(cls, subclass):
        return subclasshook(cls, subclass)
    
    def __eq__(self, other):
        """ Delegate to “compare_via_attrs(…)” """
        return compare_via_attrs(self, other)

class Enum(EnumBase):
    
    """ Base abstract processor enum. """
    __slots__ = tuple()
    
    @abstract
    def process(self, image):
        """ Process an image instance, per the processor enum instance,
            returning the processed image data.
        """
        ...
    
    @classmethod
    def __subclasshook__(cls, subclass):
        return subclasshook(cls, subclass)

class NOOp(Processor):
    
    """ A no-op processor. """
    __slots__ = tuple()
    
    def process(self, image):
        """ Return the image instance, unchanged """
        return image
    
    def __eq__(self, other):
        """ Simple type-comparison """
        return type(self) is type(other)

class Container(Processor):
    
    """ Base abstract processor container. """
    __slots__ = tuple()
    
    @classmethod
    @abstract
    def base_type(cls):
        """ Return the internal type upon which this instakit.abc.Container
            subclass is based.
        """
        ...
    
    @abstract
    def iterate(self):
        """ Return an ordered iterable of sub-processors. """
        ...
    
    @abstract
    def __len__(self): ...
    
    @abstract
    def __contains__(self, value): ...
    
    @abstract
    def __getitem__(self, idx): ...
    
    def __bool__(self):
        """ A processor container is considered Truthy if it contains values,
            and Falsey if it is empty.
        """
        return len(self) > 0
    
    def __eq__(self, other):
        """ Compare “base_type()” results and item-by-item through “iterate()” """
        if type(self).base_type() is not type(other).base_type():
            return NotImplemented
        for self_item, other_item in zip(self.iterate(),
                                         other.iterate()):
            if self_item != other_item:
                return False
        return True

class Mapping(Container):
    
    __slots__ = tuple()
    
    @abstract
    def get(self, idx, default_value): ...

class Sequence(Container):
    
    __slots__ = tuple()
    
    @abstract
    def index(self, value): ...
    
    @abstract
    def last(self): ...

class MutableContainer(Container):
    
    """ Base abstract processor mutable container. """
    __slots__ = tuple()
    
    @abstract
    def __setitem__(self, idx, value): ...
    
    @abstract
    def __delitem__(self, idx, value): ...

class MutableMapping(MutableContainer):
    
    __slots__ = tuple()
    
    @abstract
    def get(self, idx, default_value): ...
    
    @abstract
    def pop(self, idx, default_value): ...
    
    @abstract
    def update(self, iterable=None, **kwargs): ...

class MutableSequence(MutableContainer):
    
    __slots__ = tuple()
    
    @abstract
    def index(self, value): ...
    
    @abstract
    def last(self): ...
    
    @abstract
    def append(self, value): ...
    
    @abstract
    def extend(self, iterable): ...
    
    @abstract
    def pop(self, idx=-1): ...

class Fork(MutableMapping):
    
    """ Base abstract forking processor. """
    __slots__ = ('dict', '__weakref__')
    
    @classmethod
    def base_type(cls):
        return defaultdict
    
    def __init__(self, default_factory, *args, **kwargs):
        """ The `Fork` ABC implements the same `__init__(¬)` call signature as
            its delegate type, `collections.defaultdict`. A “default_factory”
            callable argument is required to fill in missing values (although
            one can pass None, which will cause a `NOOp` processor to be used).
            
            From the `collections.defaultdict` docstring:
            
           “defaultdict(default_factory[, ...]) --> dict with default factory”
            
           “The default factory is called without arguments to produce
            a new value when a key is not present, in __getitem__ only.
            A defaultdict compares equal to a dict with the same items.
            All remaining arguments are treated the same as if they were
            passed to the dict constructor, including keyword arguments.”
        
        """
        if default_factory in (None, NOOp):
            default_factory = NOOp
        if not callable(default_factory):
            raise AttributeError("Fork() requires a callable default_factory")
        
        self.dict = type(self).base_type()(default_factory, *args, **kwargs)
        super(Fork, self).__init__()
    
    @property
    def default_factory(self):
        """ The default factory for the dictionary. """
        return self.dict.default_factory
    
    @default_factory.setter
    def default_factory(self, value):
        if not callable(value):
            raise AttributeError("Fork.default_factory requires a callable value")
        self.dict.default_factory = value
    
    def __len__(self):
        """ The number of entries in the dictionary.
            See defaultdict.__len__(…) for details.
        """
        return len(self.dict)
    
    def __contains__(self, idx):
        """ True if the dictionary has the specified `idx`, else False.
            See defaultdict.__contains__(…) for details.
        """
        return idx in self.dict
    
    def __getitem__(self, idx):
        """ Get a value from the dictionary, or if no value is present,
            the return value of `default_factory()`.
            See defaultdict.__getitem__(…) for details.
        """
        return self.dict[idx]
    
    def __setitem__(self, idx, value):
        """ Set the value in the dictionary corresponding to the specified
           `idx` to the value passed, or if a value of “None” was passed,
            set the value to `instakit.abc.NOOp()` -- the no-op processor.
        """
        if value in (None, NOOp):
            value = NOOp()
        self.dict[idx] = value
    
    def __delitem__(self, idx):
        """ Delete a value from the dictionary corresponding to the specified
           `idx`, if one is present.
            See defaultdict.__delitem__(…) for details.
        """
        del self.dict[idx]
    
    def get(self, idx, default_value=None):
        """ Get a value from the dictionary, with an optional default
            value to use should a value not be present for this `idx`.
            See defaultdict.get(…) for details.
        """
        return self.dict.get(idx, default_value)
    
    def pop(self, idx, default_value=None):
        """ D.pop(idx[,d]) -> v, remove specified `idx` and return the corresponding value.
            If `idx` is not found, d is returned if given, otherwise `KeyError` is raised.
            See defaultdict.pop(…) for details.
        """
        return self.dict.pop(idx, default_value)
    
    def update(self, iterable=None, **kwargs):
        """ Update the dictionary with new key-value pairs.
            See defaultdict.update(…) for details.
        """
        self.dict.update(iterable or tuple(), **kwargs)
    
    @abstract
    def split(self, image): ...
    
    @abstract
    def compose(self, *bands): ...

class ThresholdMatrixProcessor(Processor):
    
    """ Abstract base class for a processor using a uint8 threshold matrix """
    # This is used in instakit.processors.halftone
    __slots__ = ('threshold_matrix',)
    
    LO_TUP = (0,)
    HI_TUP = (255,)
    
    def __init__(self, threshold = 128.0):
        """ Initialize with a threshold value between 0 and 255 """
        self.threshold_matrix = int(threshold)  * self.LO_TUP + \
                           (256-int(threshold)) * self.HI_TUP

class NDProcessorBase(Processor):
    
    """ An image processor ancestor class that represents PIL image
        data in a `numpy.ndarray`. This is the base abstract class,
        specifying necessary methods for subclasses to override.
        
        Note that “process(…)” has NOT been implemented yet in the
        inheritance chain – a subclass will need to furnish it.
    """
    __slots__ = tuple()
    
    @abstract
    def process_nd(self, ndimage):
        """ Override NDProcessor.process_nd(…) in subclasses
            to provide functionality that acts on image data stored
            in a `numpy.ndarray`.
        """
        ...
    
    @staticmethod
    @abstract
    def compand(ndimage): ...
    
    @staticmethod
    @abstract
    def uncompand(ndimage): ...


def test():
    """ Inline tests for instakit.abc module """
    import os
    if os.environ.get('TM_PYTHON'):
        import sys
        def print_red(text):
            print(text, file=sys.stderr)
    else:
        import colorama, termcolor
        colorama.init()
        def print_red(text):
            print(termcolor.colored(text, color='red'))
    
    import __main__
    print_red(__main__.__doc__)
    
    class SlowAtkinson(ThresholdMatrixProcessor):
        __slots__ = tuple()
        def process(self, image):
            from instakit.utils.mode import Mode
            image = Mode.L.process(image)
            for y in range(image.size[1]):
                for x in range(image.size[0]):
                    old = image.getpixel((x, y))
                    new = self.threshold_matrix[old]
                    err = (old - new) >> 3 # divide by 8.
                    image.putpixel((x, y), new)
                    for nxy in [(x+1, y),
                                (x+2, y),
                                (x-1, y+1),
                                (x, y+1),
                                (x+1, y+1),
                                (x, y+2)]:
                        try:
                            image.putpixel(nxy, int(
                            image.getpixel(nxy) + err))
                        except IndexError:
                            pass
            return image
    
    from pprint import pprint
    slow_atkinson = SlowAtkinson()
    pprint(slow_atkinson)
    print("DICT?", hasattr(slow_atkinson, '__dict__'))
    print("SLOTS?", hasattr(slow_atkinson, '__slots__'))
    pprint(slow_atkinson.__slots__)
    pprint(slow_atkinson.__class__.__base__.__slots__)
    pprint(slots_for(SlowAtkinson))
    print("THRESHOLD_MATRIX:", slow_atkinson.threshold_matrix)
    assert slow_atkinson == SlowAtkinson()
    assert NOOp() == NOOp()

if __name__ == '__main__':
    test()



#!/usr/bin/env python
# encoding: utf-8
from __future__ import print_function

def gaussian(size, sizeY=None):
    """ Returns a normalized 2D gauss kernel array for convolutions """
    from scipy import mgrid, exp
    sizeX = int(size)
    if not sizeY:
        sizeY = sizeX
    else:
        sizeY = int(sizeY)
    x, y = mgrid[-sizeX:sizeX+1,
                 -sizeY:sizeY+1]
    g = exp(-(x**2/float(sizeX)+y**2/float(sizeY)))
    return (g / g.sum()).flatten()

def gaussian_blur_kernel(ndim, kernel):
    from scipy.ndimage import convolve
    return convolve(ndim, kernel, mode='reflect')

def gaussian_blur(ndim, sigma=3, sizeY=None):
    return gaussian_blur_kernel(ndim,
                                gaussian(sigma,
                                sizeY=sizeY))

def gaussian_blur_filter(input, sigmaX=3,
                                sigmaY=3,
                                sigmaZ=0):
    from scipy.ndimage.filters import gaussian_filter
    if not sigmaY:
        sigmaY = sigmaX
    if not sigmaZ:
        sigmaZ = sigmaX
    return gaussian_filter(input, sigma=(sigmaX,
                                         sigmaY,
                                         sigmaZ), order=0,
                                                  mode='reflect')



cdef extern from "hsluv.h" nogil:
    
    void hsluv2rgb(double   h, double   s, double   l,
                   double* pr, double* pg, double* pb)
    
    void rgb2hsluv(double   r, double   g, double   b,
                   double* ph, double* ps, double* pl)
    
    void hpluv2rgb(double   h, double   s, double   l,
                   double* pr, double* pg, double* pb)
    
    void rgb2hpluv(double   r, double   g, double   b,
                   double* ph, double* ps, double* pl)



#!/usr/bin/env python
# encoding: utf-8
"""
curves.py

Adapted from this:

    https://gist.github.com/fish2000/5641c3697fa4407fcfd59099575d6938

And also this:

    https://github.com/vbalnt/filterizer/blob/master/extractCurvesFromACVFile.py

Created by FI$H 2000 on 2012-08-23.
Copyright (c) 2012-2019 Objects In Space And Time, LLC. All rights reserved.

"""
from __future__ import print_function

import numpy
import os
import struct

from PIL import Image
from enum import Enum, unique
from scipy import interpolate

from instakit.utils.static import asset
from instakit.utils.mode import Mode
from instakit.abc import Processor

interpolate_mode_strings = ('linear',
                            'nearest',
                            'zero',
                            'slinear',
                            'quadratic', 'cubic',
                            'previous', 'next',
                            'lagrange')

@unique
class InterpolateMode(Enum):
    
    # These correspond to the “kind” arg
    # from “scipy.interpolate.interp1d(…)”:
    LINEAR = 0
    NEAREST = 1
    ZERO = 2
    SLINEAR = 3
    QUADRATIC = 4
    CUBIC = 5
    PREVIOUS = 6
    NEXT = 7
    
    # This specifies LaGrange interpolation,
    # using “scipy.interpolate.lagrange(…)”:
    LAGRANGE = 8
    
    def to_string(self):
        return interpolate_mode_strings[self.value]
    
    def __str__(self):
        return self.to_string()


class SingleCurve(list):
    
    """ A SingleCurve instance is a named list of (x, y) coordinates,
        that provides programmatic access to interpolated values.
        
        It is constructed with `(name, [(x, y), (x, y)...])`; since it
        directly inherits from `__builtins__.list`, the usual methods
        e.g. `append(…)`, `insert(…)` &c. can be used to modify an
        instance of SingleCurve.
        
        Before accessing interpolated values, one first calls the
        method `interpolate(…)` with an optional argument specifying
        the interpolation mode, `mode=InterpolationMode` (q.v. the
        `InterpolationMode` enum supra.) and thereafter, instances
        of SingleCurve are callable with an x-coordinate argument,
        returning the interpolated y-coordinate.
    """
    
    def __init__(self, name, *args):
        self.name = name
        list.__init__(self, *args)
    
    def asarray(self, dtype=None):
        return numpy.array(self, dtype=dtype)
    
    def interpolate(self, mode=InterpolateMode.LAGRANGE):
        xy = self.asarray()
        if mode == InterpolateMode.LAGRANGE or mode is None:
            delegate = interpolate.lagrange(xy.T[0],
                                            xy.T[1])
        else:
            kind = InterpolateMode(mode).to_string()
            delegate = interpolate.interp1d(xy.T[0],
                                            xy.T[1], kind=kind)
        self.delegate = delegate
        return self
    
    def __call__(self, value):
        if not hasattr(self, 'delegate'):
            self.interpolate()
        delegate = self.delegate
        return delegate(value)


class CurveSet(Processor):
    
    """ A CurveSet instance represents an ACV file, as generated by the
        Adobe® Photoshop™ application, whose data encodes a set of
        image-adjustment curves.
        
        The simplest use is to read an existing set of curves from an
        existant ACV file; one instantiates a CurveSet like so:
        
            mycurveset = CurveSet('path/to/curveset.acv')
        
        …one can then use `mycurveset.process(…)` to process PIL images,
        or one can access underlying curve data via `mycurveset.curves`;
        subsequently the curveset can be rewritten to a new ACV file
        with `mycurveset.write_acv(acv_file_path)`.
    """
    __slots__ = ('_is_builtin', 'count', 'curves',
                                'path',  'name',
                                'interpolation_mode')
    
    acv = 'acv'
    dotacv = '.' + acv
    channels = ('composite', 'red', 'green', 'blue')
    valid_modes = ( Mode.RGB, Mode.MONO, Mode.L )
    
    @classmethod
    def builtin(cls, name):
        print("Reading curves [builtin] %s%s" % (name, cls.dotacv))
        acv_path = asset.path(cls.acv, "%s%s" % (name, cls.dotacv))
        out = cls(acv_path)
        out._is_builtin = True
        return out
    
    @classmethod
    def instakit_names(cls):
        return [curve_file.rstrip(cls.dotacv) \
            for curve_file in asset.listfiles(cls.acv) \
            if curve_file.lower().endswith(cls.dotacv)]
    
    @classmethod
    def instakit_curve_sets(cls):
        return [cls.builtin(name) for name in cls.instakit_names()]
    
    @classmethod
    def channel_name(cls, idx):
        try:
            return cls.channels[idx]
        except IndexError:
            return "channel%s" % idx
    
    def __init__(self, path, interpolation_mode=InterpolateMode.LAGRANGE):
        self.count = 0
        self.curves = []
        self._is_builtin = False
        self.path = os.path.abspath(path)
        self.name = os.path.basename(path)
        self.interpolation_mode = interpolation_mode
        if os.path.isfile(self.path):
            self.read_acv(self.path,
                          self.interpolation_mode)
    
    @property
    def is_builtin(self):
        return self._is_builtin
    
    @property
    def file_exists(self):
        return os.path.isfile(self.path)
    
    @staticmethod
    def read_one_curve(acv_file, name, interpolation_mode):
        curve = SingleCurve(name)
        points_in_curve, = struct.unpack("!h", acv_file.read(2))
        for _ in range(points_in_curve):
            y, x = struct.unpack("!hh", acv_file.read(4))
            curve.append((x, y))
        return curve.interpolate(interpolation_mode)
    
    @staticmethod
    def write_one_curve(acv_file, curve):
        points_in_curve = len(curve)
        acv_file.write(struct.pack("!h", points_in_curve))
        for idx in range(points_in_curve):
            x, y = curve[idx]
            acv_file.write(struct.pack("!hh", y, x))
        return points_in_curve
    
    def read_acv(self, acv_path, interpolation_mode):
        if not self.file_exists:
            raise IOError("Can't read nonexistant ACV file: %s" % self.path)
        with open(acv_path, "rb") as acv_file:
            _, self.count = struct.unpack("!hh", acv_file.read(4))
            for idx in range(self.count):
                self.curves.append(
                    self.read_one_curve(acv_file,
                                   type(self).channel_name(idx),
                                        interpolation_mode))
    
    def write_acv(self, acv_path):
        if self.count < 1:
            raise ValueError("Can't write empty curveset as ACV data")
        with open(acv_path, "wb") as acv_file:
            acv_file.write(struct.pack("!hh", 0, self.count))
            for curve in self.curves:
                self.write_one_curve(acv_file, curve)
    
    def process(self, image):
        mode = Mode.of(image)
        if mode not in type(self).valid_modes:
            image = Mode.RGB.process(image)
        elif mode is not Mode.RGB:
            return Image.eval(Mode.L.process(image),
                              self.curves[0])
        # The image to be RGB-modes at this point:
        adjusted_bands = []
        for idx, band in enumerate(image.split()):
            adjusted_bands.append(
                Image.eval(band,
                           lambda v: self.curves[idx+1](v)))
        return Mode.RGB.merge(*adjusted_bands)
    
    def add(self, curve):
        self.curves.append(curve)
        self.count = len(self.curves)
    
    def __repr__(self):
        cls_name = getattr(type(self), '__qualname__',
                   getattr(type(self), '__name__'))
        address = id(self)
        label = self.is_builtin and '[builtin]' or self.name
        interp = self.interpolation_mode or InterpolateMode.LAGRANGE
        parenthetical = "%s, %d, %s" % (label, self.count, interp)
        return "%s(%s) @ <%s>" % (cls_name, parenthetical, address)


def test():
    curve_sets = CurveSet.instakit_curve_sets()
    
    image_paths = list(map(
        lambda image_file: asset.path('img', image_file),
            asset.listfiles('img')))
    image_inputs = list(map(
        lambda image_path: Mode.RGB.open(image_path),
            image_paths))
    
    for image_input in image_inputs[:1]:
        image_input.show()
        for curve_set in curve_sets:
            curve_set.process(image_input).show()
    
    print(curve_sets)
    print(image_paths)
    
    import tempfile
    temppath = tempfile.mktemp(suffix='.acv')
    assert not CurveSet(path=temppath).file_exists
    
if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
from __future__ import print_function
from math import fabs, pow as mpow

from instakit.utils.mode import Mode
from instakit.abc import Processor

PERCENT_ADMONISHMENT = "Do you not know how percents work??!"

def gcr(image, percentage=20, revert_mode=False):
    ''' basic “Gray Component Replacement” function. Returns a CMYK image* with 
        percentage gray component removed from the CMY channels and put in the
        K channel, e.g. for percentage=100, (41, 100, 255, 0) >> (0, 59, 214, 41).
        
    {*} This is the default behavior – to return an image of the same mode as that
        of which was originally provided, pass the value for the (optional) keyword
        argument `revert_mode` as `True`.
    '''
    # from http://stackoverflow.com/questions/10572274/halftone-images-in-python
    
    if percentage is None:
        return revert_mode and image or Mode.CMYK.process(image)
    
    if percentage > 100 or percentage < 1:
        raise ValueError(PERCENT_ADMONISHMENT)
    
    percent = percentage / 100
    
    original_mode = Mode.of(image)
    cmyk_channels = Mode.CMYK.process(image).split()
    width, height = image.size
    
    cmyk_image = []
    for channel in cmyk_channels:
        cmyk_image.append(channel.load())
    
    for x in range(width):
        for y in range(height):
            gray = int(min(cmyk_image[0][x, y],
                           cmyk_image[1][x, y],
                           cmyk_image[2][x, y]) * percent)
            cmyk_image[0][x, y] -= gray
            cmyk_image[1][x, y] -= gray
            cmyk_image[2][x, y] -= gray
            cmyk_image[3][x, y] = gray
    
    recomposed = Mode.CMYK.merge(*cmyk_channels)
    
    if revert_mode:
        return original_mode.process(recomposed)
    return recomposed

class BasicGCR(Processor):
    
    __slots__ = ('percentage', 'revert_mode')
    __doc__ = gcr.__doc__
    
    def __init__(self, percentage=20, revert_mode=False):
        if percentage is None:
            raise ValueError(PERCENT_ADMONISHMENT)
        if percentage > 100 or percentage < 1:
            raise ValueError(PERCENT_ADMONISHMENT)
        self.percentage = percentage
        self.revert_mode = revert_mode
    
    def process(self, image):
        return gcr(image, percentage=self.percentage,
                          revert_mode=self.revert_mode)

def hex2rgb(h):
    """ Convert a hex string or number to an RGB triple """
    # q.v. https://git.io/fh9E2
    if isinstance(h, str):
        return hex2rgb(int(h[1:] if h.startswith('#') else h, 16))
    return (h >> 16) & 0xff, (h >> 8) & 0xff, h & 0xff

def compand(v):
    """ Compand a linearized value to an sRGB byte value """
    # q.v. http://www.brucelindbloom.com/index.html?Math.html
    V = (v <= 0.0031308) and (v * 12.92) or fabs((1.055 * mpow(v, 1 / 2.4)) - 0.055)
    return int(V * 255.0)

def uncompand(A):
    """ Uncompand an sRGB byte value to a linearized value """
    # q.v. http://www.brucelindbloom.com/index.html?Eqn_RGB_to_XYZ.html
    V = A / 255.0
    return (V <= 0.04045) and (V / 12.92) or mpow(((V + 0.055) / 1.055), 2.4)

def ucr(image, revert_mode=False):
    ''' basic “Under-Color Removal” function. Returns a CMYK image* in which regions
        containing overlapping C, M, and Y ink are replaced with K (“Key”, née “BlacK”)
        ink. Images are first converted to RGB and linearized in order to perform the UCR
        operation in linear space. E.g.:
        
        0xFFDE17 >  rgb(255, 222, 23)
                 >  RGB(1.0, 0.7304607400903537, 0.008568125618069307)
                 >  CMY(0.0, 0.26953925990964633, 0.9914318743819307)
        
        rgb() > RGB() > CMY() > CMYK(41, 100, 255, 0) >> cmyk(0, 59, 214, 41).
        
    {*} This is the default behavior – to return an image of the same mode as that
        of which was originally provided, pass the value for the (optional) keyword
        argument `revert_mode` as `True`.
    '''
    # Adapted from http://www.easyrgb.com/en/math.php#text13
    # N.B. this is not, out of the gate, particularly well-optimized
    
    original_mode = Mode.of(image)
    width, height = image.size
    cmyk_target = Mode.CMYK.new(image.size, color=0)
    rgb_channels = Mode.RGB.process(image).split()
    cmyk_channels = cmyk_target.split()
    
    rgb_image = []
    for channel in rgb_channels:
        rgb_image.append(channel.load())
    
    cmyk_image = []
    for channel in cmyk_channels:
        cmyk_image.append(channel.load())
    
    for x in range(width):
        for y in range(height):
            # Get the rgb byte values:
            rgb = (rgb_image[0][x, y],
                   rgb_image[1][x, y],
                   rgb_image[2][x, y])
            
            # Uncompand rgb bytes to linearized RGB:
            RGB = (uncompand(v) for v in rgb)
            
            # Convert linear RGB to linear CMY:
            (C, M, Y) = (1.0 - V for V in RGB)
            
            # Perform simple UCR with the most combined
            # overlapping C/M/Y ink values:
            K = min(C, M, Y, 1.0)
            
            if K == 1:
                C = M = Y = 0
            else:
                denominator = (1 - K)
                C = (C - K) / denominator
                M = (M - K) / denominator
                Y = (Y - K) / denominator
            
            # Recompand linear CMYK to cmyk byte values for Pillow:
            cmyk_image[0][x, y] = compand(C)
            cmyk_image[1][x, y] = compand(M)
            cmyk_image[2][x, y] = compand(Y)
            cmyk_image[3][x, y] = compand(K)
    
    recomposed = Mode.CMYK.merge(*cmyk_channels)
    
    if revert_mode:
        return original_mode.process(recomposed)
    return recomposed

class BasicUCR(Processor):
    
    __slots__ = ('revert_mode',)
    __doc__ = ucr.__doc__
    
    def __init__(self, revert_mode=False):
        self.revert_mode = revert_mode
    
    def process(self, image):
        return ucr(image, revert_mode=self.revert_mode)

class DemoUCR(object):
    
    """ Demonstrate each phase of the UCR color-conversion process """
    
    def __init__(self, hex_triple):
        """ Initialize with a hex-encoded RGB triple, either as a string
            or as a hexadecimal integer, e.g.:
                
                >>> onedemo = DemoUCR(0xD8DCAB)
                >>> another = DemoUCR('#6A9391')
        """
        self.hex_triple = hex_triple
    
    def calculate(self):
        self.rgb = self.get_rgb()
        self.RGB = self.get_RGB()
        self.CMY = self.get_CMY()
        self.CMYK = self.get_CMYK()
        self.cmyk = self.get_cmyk()
    
    def get_rgb(self):
        """ Return the rgb byte-value (0-255) 3-tuple corresponding to the
            initial hex value
        """
        return hex2rgb(self.hex_triple)
    
    def get_RGB(self):
        """ Return the linearized (uncompanded) RGB 3-tuple version of the
            rgb 3-byte value tuple
        """
        return tuple(uncompand(v) for v in self.rgb)
    
    def get_CMY(self):
        """ Return the linearized (uncompanded) CMY color-model analog of the
            linear RGB value 3-tuple
        """
        (C, M, Y) = (1.0 - V for V in self.RGB)
        return (C, M, Y)
    
    def get_CMYK(self):
        """ Return the UCR’ed -- the under-color removed -- linearized CMYK analog
            of the linear CMY color-model value tuple
        """
        K = min(*self.CMY, 1.0)
        (C, M, Y) = self.CMY
        
        if K == 1:
            C = M = Y = 0
        else:
            denominator = (1 - K)
            C = (C - K) / denominator
            M = (M - K) / denominator
            Y = (Y - K) / denominator
        return (C, M, Y, K)
    
    def get_cmyk(self):
        """ Return the companded cmyk byte-value (0-255) 4-tuple Pillow-friendly
            CMYK analog of the initial value
        """
        return tuple(compand(V) for V in self.CMYK)
    
    def stringify_demo_values(self):
        """ Return each of the values, step-by-step in the conversion process,
            admirably formatted in a handsome manner suitable for printing
        """
        self.calculate()
        return """
            
    %(hex_triple)s > rgb%(rgb)s
            > RGB%(RGB)s
            > CMY%(CMY)s
            > CMYK%(CMYK)s
            > cmyk%(cmyk)s
        
        """ % self.__dict__
    
    def __str__(self):
        """ Stringify yo self """
        return self.stringify_demo_values()


def test():
    from instakit.utils.static import asset
    from itertools import chain
    from os.path import relpath
    from pprint import pprint
    
    start = "/usr/local/Cellar/python/3.7.2_2/Frameworks/Python.framework/Versions/3.7/lib/python3.7/site-packages"
    
    image_paths = list(map(lambda image_file: asset.path('img', image_file), asset.listfiles('img')))
    image_inputs = list(map(lambda image_path: Mode.RGB.open(image_path), image_paths))
    # print("len:", len(list(image_paths)), len(list(image_inputs)), len(list(callables)))
    
    print('\t<<<<<<<<<<<<<<<------------------------------------------------------>>>>>>>>>>>>>>>')
    print()
    
    functions = (gcr, ucr)
    processors = (BasicGCR(), BasicUCR(), Mode.CMYK)
    callables = chain((processor.process for processor in processors), functions)
    
    image_components = zip(image_paths, image_inputs, callables)
    
    for path, image, process_functor in image_components:
        print("«TESTING: %s»" % relpath(path, start=start))
        print()
        tup = image.size + (image.mode,)
        print("¬ Input: %sx%s %s" % tup)
        print("¬ Calling functor on image…")
        result = process_functor(image)
        tup = result.size + (result.mode,)
        print("¬ Output: %sx%s %s" % tup)
        print("¬ Displaying…")
        print()
        result.show()
    
    print("«¡SUCCESS!»")
    print()
    
    print("«TESTING: MANUAL CALLABLES»")
    # print()
    
    if len(image_inputs):
        image = image_inputs.pop()
        
        # Test GCR function:
        gcred = gcr(image)
        assert gcred.mode == Mode.CMYK.value.mode
        assert Mode.of(gcred) is Mode.CMYK
        # gcred.show()
        
        # close image:
        image.close()
    
    if len(image_inputs):
        image = image_inputs.pop()
        
        # Test UCR function:
        ucred = ucr(image)
        assert ucred.mode == Mode.CMYK.value.mode
        assert Mode.of(ucred) is Mode.CMYK
        # ucred.show()
        
        # close image:
        image.close()
    
    if len(image_inputs):
        image = image_inputs.pop()
        
        # Test GCR processor:
        gcr_processor = BasicGCR()
        gcred = gcr_processor.process(image)
        assert gcred.mode == Mode.CMYK.value.mode
        assert Mode.of(gcred) is Mode.CMYK
        # gcred.show()
        
        # close image:
        image.close()
    
    if len(image_inputs):
        image = image_inputs.pop()
        
        # Test UCR processor:
        ucr_processor = BasicUCR()
        ucred = ucr_processor.process(image)
        assert ucred.mode == Mode.CMYK.value.mode
        assert Mode.of(ucred) is Mode.CMYK
        # ucred.show()
        
        # close image:
        image.close()
    
    print("«¡SUCCESS!»")
    print()
    
    print('\t<<<<<<<<<<<<<<<------------------------------------------------------>>>>>>>>>>>>>>>')
    print()
    
    pprint(list(relpath(path, start=start) for path in image_paths))
    print()
    
    print("«TESTING: DemoUCR ALGORITHM-STAGE TRACE PRINTER»")
    print()
    
    print(DemoUCR("#BB2F53"))
    print()
    
    print(DemoUCR(0x6F2039))
    print()


if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
from __future__ import print_function

import argparse
import enum
import importlib
import inspect
import types

from pprint import pprint

class Parameter(object):
    """ A placeholder object, used for the moment in the inline tests """
    pass

QUALIFIER = '.'

def dotpath_join(base, *addenda):
    """ Join dotpath elements together as one, á la os.path.join(…) """
    for addendum in addenda:
        if not base.endswith(QUALIFIER):
            base += QUALIFIER
        if addendum.startswith(QUALIFIER):
            if len(addendum) == 1:
                raise ValueError('operand too short: %s' % addendum)
            addendum = addendum[1:]
        base += addendum
    # N.B. this might be overthinking it -- 
    # maybe we *want* to allow dotpaths
    # that happen to start and/or end with dots?
    if base.endswith(QUALIFIER):
        return base[:-1]
    return base

def qualified_import(qualified):
    """ Import a qualified thing-name.
        e.g. 'instakit.processors.halftone.FloydSteinberg'
    """
    if QUALIFIER not in qualified:
        raise ValueError("qualified_import() needs a qualified name "
                         "(got %s)" % qualified)
    head = qualified.split(QUALIFIER)[-1]
    tail = qualified.replace("%s%s" % (QUALIFIER, head), '')
    module = importlib.import_module(tail)
    cls = getattr(module, head)
    print("Qualified Import: %s" % qualified)
    return cls

def qualified_name_tuple(cls):
    """ Get the module name and the thing-name for a class.
        e.g. ('instakit.processors.halftone', 'FloydSteinberg')
    """
    mod_name = getattr(cls, '__module__')
    cls_name = getattr(cls, '__qualname__',
               getattr(cls, '__name__'))
    return mod_name, cls_name

def qualified_name(cls):
    """ Get a qualified thing-name for a class.
        e.g. 'instakit.processors.halftone.FloydSteinberg'
    """
    mod_name, cls_name = qualified_name_tuple(cls)
    out = "%s%s%s" % (mod_name, QUALIFIER, cls_name)
    print("Qualified Name: %s" % out)
    return out

class Nothing(object):
    """ Placeholder singleton, signifying nothing """
    __slots__ = tuple()
    def __new__(cls, *a, **k):
        return Nothing

def check_parameter_default(param_default):
    """ Filter result values coming from inspect.signature(…) """
    if param_default == inspect._empty:
        return Nothing
    return param_default

def default_arguments(cls):
    """ Get a dictionary of the keyword arguments with provided defaults,
        as furnished by a given classes’ “__init__” function.
    """
    try:
        signature = inspect.signature(cls)
    except (ValueError, TypeError) as exc:
        m, n = qualified_name_tuple(cls)
        qn = "%s%sSlow%s" % (m.replace('ext.', ''), QUALIFIER, n) # WTF HAX
        NonCythonCls = qualified_import(qn)
        if qualified_name(NonCythonCls) != qualified_name(cls):
            return default_arguments(NonCythonCls)
        else:
            raise exc
    if len(signature.parameters) < 1:
        return {}
    return { parameter.name : check_parameter_default(parameter.default) \
                                for parameter \
                                in signature.parameters.values() }

def is_enum(cls):
    """ Predicate function to ascertain whether a class is an Enum. """
    return enum.Enum in cls.__mro__

def enum_choices(cls):
    """ Return a list of the names of the given Enum class members. """
    return [choice.name for choice in cls]

FILE_ARGUMENT_NAMES = ('path', 'pth', 'file')

def add_argparser(subparsers, cls):
    """ Add a subparser -- an instance of “argparse.ArgumentParser” --
        with arguments and defaults matching the keyword arguments and
        defaults provided by the given class (q.v. “default_arguments(…)”
        definition supra.)
    """
    qualname = qualified_name(cls)
    cls_help = getattr(cls, '__doc__', None) or "help for %s" % qualname
    parser = subparsers.add_parser(qualname, help=cls_help)
    if is_enum(cls): # Deal with enums
        argument_name = cls.__name__.lower()
        add_argument_args = dict(choices=enum_choices(cls),
                                 type=str,
                                 help='help for enum %s' % argument_name)
        parser.add_argument(argument_name,
                          **add_argument_args)
    else: # Deal with __init__ signature
        for argument_name, argument_value in default_arguments(cls).items():
            argument_type = type(argument_value)
            argument_required = False
            add_argument_args = dict(help='help for argument %s' % argument_name)
            if argument_value is not Nothing:
                add_argument_args.update({ 'default' : argument_value })
            else:
                add_argument_args.update({ 'type' : argument_name in FILE_ARGUMENT_NAMES \
                                                and argparse.FileType('rb') \
                                                 or str })
                argument_required = True
            if argument_type is bool:
                add_argument_args.update({ 'action' : 'store_true' })
            elif argument_type is type(None):
                add_argument_args.update({ 'type' : str })
            elif is_enum(argument_type):
                add_argument_args.update({ 'choices' : enum_choices(argument_type),
                                              'type' : str })
            argument_template = argument_required and '%s' or '--%s'
            parser.add_argument(argument_template % argument_name,
                              **add_argument_args)
    return parser

functype = types.FunctionType

def get_processors_from(module_name):
    """ Memoized processor-extraction function """
    from instakit.utils.static import asset
    if not hasattr(get_processors_from, 'cache'):
        get_processors_from.cache = {}
    if module_name not in get_processors_from.cache:
        processors = []
        module = importlib.import_module(module_name)
        print("Module: %s (%s)" % (module.__name__,
                    asset.relative(module.__file__)))
        for thing in (getattr(module, name) for name in dir(module)):
            if hasattr(thing, 'process'):
                print("Found thing: %s" % thing)
                if module.__name__ in thing.__module__:
                    if thing not in processors:
                        if type(getattr(thing, 'process')) is functype:
                            processors.append(thing)
        get_processors_from.cache[module_name] = tuple(processors)
    return get_processors_from.cache[module_name] 


def test():
    
    # Test “qualified_import()”:
    print("Testing “qualified_import()”…")
    
    class_name = 'instakit.processors.halftone.SlowFloydSteinberg'
    ImportedFloydSteinberg = qualified_import(class_name)
    assert ImportedFloydSteinberg.__name__ == 'SlowFloydSteinberg'
    assert ImportedFloydSteinberg.__qualname__ == 'SlowFloydSteinberg'
    assert ImportedFloydSteinberg.__module__ == 'instakit.processors.halftone'
    
    class_name = 'instakit.processors.halftone.Atkinson' # TWIST!!
    ImportedAtkinson = qualified_import(class_name)
    assert ImportedAtkinson.__name__ == 'Atkinson'
    assert ImportedAtkinson.__qualname__ == 'Atkinson'
    assert ImportedAtkinson.__module__ == 'instakit.processors.ext.halftone'
    
    print("Success!")
    print()
    
    # Test “qualified_name()”:
    print("Testing “qualified_name()”…")
    
    class_name = qualified_name(Parameter)
    assert class_name == '__main__.Parameter'
    
    class_name = qualified_name(ImportedFloydSteinberg)
    assert class_name == 'instakit.processors.halftone.SlowFloydSteinberg'
    
    class_name = qualified_name(ImportedAtkinson)
    assert class_name == 'instakit.processors.ext.halftone.Atkinson'
    
    print("Success!")
    print()
    
    # Test “Nothing”:
    print("Testing “Nothing”…")
    
    assert type(Nothing) == type
    assert Nothing() == Nothing
    
    print("Success!")
    print()
    
    # Test “default_arguments()”:
    print("Testing “default_arguments()”…")
    
    default_args = default_arguments(ImportedFloydSteinberg)
    assert default_args == dict(threshold=128.0)
    
    slow_atkinson = 'instakit.processors.halftone.SlowAtkinson'
    default_args = default_arguments(qualified_import(slow_atkinson))
    assert default_args == dict(threshold=128.0)
    
    noise = 'instakit.processors.noise.GaussianNoise'
    default_args = default_arguments(qualified_import(noise))
    assert default_args == dict()
    
    contrast = 'instakit.processors.adjust.Contrast'
    default_args = default_arguments(qualified_import(contrast))
    assert default_args == dict(value=1.0)
    
    unsharp_mask = 'instakit.processors.blur.UnsharpMask'
    default_args = default_arguments(qualified_import(unsharp_mask))
    assert default_args == dict(radius=2,
                                percent=150,
                                threshold=3)
    
    curves = 'instakit.processors.curves'
    curveset = dotpath_join(curves, 'CurveSet')
    interpolate_mode = dotpath_join(curves, 'InterpolateMode')
    ImportedInterpolateMode = qualified_import(interpolate_mode)
    default_args = default_arguments(qualified_import(curveset))
    LAGRANGE = ImportedInterpolateMode.LAGRANGE
    assert default_args == dict(path=Nothing,
                                interpolation_mode=LAGRANGE)
    
    print("Success!")
    print()
    
    # Test “is_enum()”:
    print("Testing “is_enum()”…")
    
    assert is_enum(ImportedInterpolateMode)
    assert not is_enum(Parameter)
    assert not is_enum(ImportedFloydSteinberg)
    assert not is_enum(ImportedAtkinson)
    
    mode = 'instakit.utils.mode.Mode'
    assert is_enum(qualified_import(mode))
    
    interpolate_mode = 'instakit.processors.curves.InterpolateMode'
    assert is_enum(qualified_import(interpolate_mode))
    
    noise_mode = 'instakit.processors.noise.NoiseMode'
    assert is_enum(qualified_import(noise_mode))
    
    print("Success!")
    print()
    
    # Test “add_argparser()”:
    print("Testing “add_argparser()”…")
    
    parser = argparse.ArgumentParser(prog='instaprocess',
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    
    parser.add_argument('--verbose', '-v',
                        action='store_true',
                        help="print verbose messages to STDOUT")
    
    processor_names = ('adjust', 'blur', 'curves', 'halftone', 'noise', 'squarecrop')
    utility_names = ('colortype', 'filesystem', 'gcr', 'kernels', 'lutmap',
                     'misc', 'mode', 'ndarrays', 'pipeline', 'static', 'stats')
    
    module_names = []
    module_names.extend(['instakit.processors.%s' % name for name in processor_names])
    module_names.extend(['instakit.utils.%s' % name for name in utility_names])
    
    processors = {}
    
    for module_name in module_names:
        processors[module_name] = get_processors_from(module_name)
    
    subparsers = parser.add_subparsers(help="subcommands for instakit processors")
    for processor_tuple in processors.values():
        for processor in processor_tuple:
            add_argparser(subparsers, processor)
    
    pprint(processors, indent=4)
    
    print()
    ns = parser.parse_args(['-h'])
    print(ns)
    
    print()
    ns = parser.parse_args(['instakit.utils.mode.Mode', '--help'])
    print(ns)
    
    print("Success!")
    print()


if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
"""
halftone.py

Created by FI$H 2000 on 2012-08-23.
Copyright (c) 2012 Objects In Space And Time, LLC. All rights reserved.
"""
from __future__ import print_function

from PIL import ImageDraw

from instakit.utils import pipeline, gcr
from instakit.utils.mode import Mode
from instakit.utils.stats import histogram_mean
from instakit.abc import Processor, ThresholdMatrixProcessor


class SlowAtkinson(ThresholdMatrixProcessor):
    
    """ It’s not a joke, this processor is slow as fuck;
        if at all possible, use the cythonized version instead
        (q.v. instakit.processors.ext.Atkinson) and never ever
        use this one if at all possible – unless, like, you’re
        being paid by the hour or somesuch. Up to you dogg.
    """
    __slots__ = tuple()
    
    def process(self, image):
        """ The process call returns a monochrome ('L'-mode) image """
        image = Mode.L.process(image)
        for y in range(image.size[1]):
            for x in range(image.size[0]):
                old = image.getpixel((x, y))
                new = self.threshold_matrix[old]
                err = (old - new) >> 3 # divide by 8.
                image.putpixel((x, y), new)
                for nxy in [(x+1, y),
                            (x+2, y),
                            (x-1, y+1),
                            (x, y+1),
                            (x+1, y+1),
                            (x, y+2)]:
                    try:
                        image.putpixel(nxy, int(
                        image.getpixel(nxy) + err))
                    except IndexError:
                        pass # it happens, evidently.
        return image

class SlowFloydSteinberg(ThresholdMatrixProcessor):
    
    """ A similarly super-slow reference implementation of Floyd-Steinberg.
        Adapted from an RGB version here: https://github.com/trimailov/qwer
    """
    __slots__ = tuple()
    
    # Precalculate fractional error multipliers:
    SEVEN_FRAC = 7/16
    THREE_FRAC = 3/16
    CINCO_FRAC = 5/16
    ALONE_FRAC = 1/16
    
    def process(self, image):
        """ The process call returns a monochrome ('L'-mode) image """
        # N.B. We store local references to the fractional error multipliers
        # to avoid the Python internal-dict-stuff member-lookup overhead:
        image = Mode.L.process(image)
        SEVEN_FRAC = type(self).SEVEN_FRAC
        THREE_FRAC = type(self).THREE_FRAC
        CINCO_FRAC = type(self).CINCO_FRAC
        ALONE_FRAC = type(self).ALONE_FRAC
        for y in range(image.size[1]):
            for x in range(image.size[0]):
                old = image.getpixel((x, y))
                new = self.threshold_matrix[old]
                image.putpixel((x, y), new)
                err = old - new
                for nxy in [((x+1, y),      SEVEN_FRAC),
                            ((x-1, y+1),    THREE_FRAC),
                            ((x, y+1),      CINCO_FRAC),
                            ((x+1, y+1),    ALONE_FRAC)]:
                    try:
                        image.putpixel(nxy[0], int(
                        image.getpixel(nxy[0]) + err * nxy[1]))
                    except IndexError:
                        pass # it happens, evidently.
        return image

# Register the stub as a instakit.abc.Processor “virtual subclass”:
@Processor.register
class Problematic(object):
    def __init__(self):
        raise TypeError("Fast-math version couldn't be imported")

try:
    # My man, fast Bill Atkinson
    from instakit.processors.ext.halftone import Atkinson as FastAtkinson
except ImportError:
    Atkinson = SlowAtkinson
    FastAtkinson = Problematic
else:
    # Register the Cythonized processor with the ABC:
    Atkinson = Processor.register(FastAtkinson)

try:
    # THE FLOYDSTER
    from instakit.processors.ext.halftone import FloydSteinberg as FastFloydSteinberg
except ImportError:
    FloydSteinberg = SlowFloydSteinberg
    FastFloydSteinberg = Problematic
else:
    # Register the Cythonized processor with the ABC:
    FloydSteinberg = Processor.register(FastFloydSteinberg)

class CMYKAtkinson(Processor):
    
    """ Create a full-color CMYK Atkinson-dithered halftone, with gray-component
        replacement (GCR) at a specified percentage level
    """
    __slots__ = ('gcr', 'overprinter')
    
    def __init__(self, gcr=20):
        self.gcr = max(min(100, gcr), 0)
        self.overprinter = pipeline.BandFork(Atkinson, mode='CMYK')
    
    def process(self, image):
        return pipeline.Pipe(gcr.BasicGCR(self.gcr),
                                          self.overprinter).process(image)

class CMYKFloydsterBill(Processor):
    
    """ Create a full-color CMYK Atkinson-dithered halftone, with gray-component
        replacement (GCR) and OH SHIT SON WHAT IS THAT ON THE CYAN CHANNEL DOGG
    """
    __slots__ = ('gcr', 'overprinter')
    
    def __init__(self, gcr=20):
        self.gcr = max(min(100, gcr), 0)
        self.overprinter = pipeline.BandFork(Atkinson, mode='CMYK')
        self.overprinter.update({ 'C' : SlowFloydSteinberg() })
    
    def process(self, image):
        return pipeline.Pipe(gcr.BasicGCR(self.gcr),
                                          self.overprinter).process(image)

class DotScreen(Processor):
    
    """ This processor creates a monochrome dot-screen halftone pattern
        from an image. While this may be useful on its own, it is far
        more useful when used across all channels of a CMYK image in
        a BandFork or OverprintFork processor operation (q.v. sources
        of `instakit.utils.pipeline.BandFork` et al. supra.) serially,
        with either a gray-component replacement (GCR) or an under-color
        replacement (UCR) function.
        
        Regarding the latter two operations, instakit only has a basic
        GCR implementation currently, at the time of writing – q.v. the
        `instakit.utils.gcr` module sub.
        
        Adapted originally from this sample code:
            https://stackoverflow.com/a/10575940/298171
    """
    __slots__ = ('sample', 'scale', 'angle')
    
    def __init__(self, sample=1, scale=2, angle=0):
        self.sample = sample
        self.scale = scale
        self.angle = angle
    
    def process(self, image):
        orig_width, orig_height = image.size
        image = Mode.L.process(image).rotate(self.angle, expand=1)
        width, height = image.size
        halftone = Mode.L.new((width * self.scale,
                              height * self.scale))
        dotscreen = ImageDraw.Draw(halftone)
        
        SAMPLE = self.sample
        SCALE = self.scale
        ANGLE = self.angle
        
        for y in range(0, height, SAMPLE):
            for x in range(0, width, SAMPLE):
                cropbox = image.crop((x,          y,
                                      x + SAMPLE, y + SAMPLE))
                diameter = (histogram_mean(cropbox) / 255) ** 0.5
                edge = 0.5 * (1 - diameter)
                xpos, ypos = (x + edge) * SCALE, (y + edge) * SCALE
                boxedge = SAMPLE * diameter * SCALE
                dotscreen.ellipse((xpos,           ypos,
                                   xpos + boxedge, ypos + boxedge),
                                   fill=255)
        
        halftone = halftone.rotate(-ANGLE, expand=1)
        tone_width, tone_height = halftone.size
        xx = (tone_width  - orig_width  * SCALE) / 2
        yy = (tone_height - orig_height * SCALE) / 2
        return halftone.crop((xx,                      yy,
                              xx + orig_width * SCALE, yy + orig_height * SCALE))

class CMYKDotScreen(Processor):
    
    """ Create a full-color CMYK dot-screen halftone, with gray-component
        replacement (GCR), individual rotation angles for each channel’s
        dot-screen, and resampling value controls.
    """
    __slots__ = ('overprinter', 'sample', 'scale')
    
    def __init__(self,      gcr=20,
                 sample=10, scale=10,
                  thetaC=0, thetaM=15, thetaY=30, thetaK=45):
        """ Initialize an internal instakit.utils.pipeline.OverprintFork() """
        self.sample = sample
        self.scale = scale
        self.overprinter = pipeline.OverprintFork(None, gcr=gcr)
        self.overprinter['C'] = DotScreen(angle=thetaC, sample=sample, scale=scale)
        self.overprinter['M'] = DotScreen(angle=thetaM, sample=sample, scale=scale)
        self.overprinter['Y'] = DotScreen(angle=thetaY, sample=sample, scale=scale)
        self.overprinter['K'] = DotScreen(angle=thetaK, sample=sample, scale=scale)
        self.overprinter.apply_CMYK_inks()
    
    @property
    def gcr_percentage(self):
        return self.overprinter.basicgcr.percentage
    
    def angle(self, band_label):
        if band_label not in self.overprinter.band_labels:
            raise ValueError('invalid band label')
        return self.overprinter[band_label].angle
    
    @property
    def thetaC(self):
        """ Return the C-band halftone screen’s rotation """
        return self.angle('C')
    
    @property
    def thetaM(self):
        """ Return the M-band halftone screen’s rotation """
        return self.angle('M')
    
    @property
    def thetaY(self):
        """ Return the Y-band halftone screen’s rotation """
        return self.angle('Y')
    
    @property
    def thetaK(self):
        """ Return the K-band halftone screen’s rotation """
        return self.angle('K')
    
    def process(self, image):
        return self.overprinter.process(image)


if __name__ == '__main__':
    from instakit.utils.static import asset
    
    image_paths = list(map(
        lambda image_file: asset.path('img', image_file),
            asset.listfiles('img')))
    image_inputs = list(map(
        lambda image_path: Mode.RGB.open(image_path),
            image_paths))
    
    for image_input in image_inputs:
        # image_input.show()
        
        # Atkinson(threshold=128.0).process(image_input).show()
        # FloydSteinberg(threshold=128.0).process(image_input).show()
        # SlowFloydSteinberg(threshold=128.0).process(image_input).show()
        
        CMYKAtkinson().process(image_input).show()
        # CMYKFloydsterBill().process(image_input).show()
        CMYKDotScreen(sample=10, scale=4).process(image_input).show()
    
    print(image_paths)



#!/usr/bin/env python
# encoding: utf-8
from __future__ import print_function

import collections.abc
from functools import wraps

try:
    import six
except ImportError:
    class FakeSix(object):
        @property
        def string_types(self):
            return tuple()
    six = FakeSix()

UTF8_ENCODING = 'UTF-8'

def tuplize(*items):
    """ Return a new tuple containing all non-`None` arguments """
    return tuple(item for item in items if item is not None)

def uniquify(*items):
    """ Return a tuple with a unique set of all non-`None` arguments """
    return tuple(frozenset(item for item in items if item is not None))

def listify(*items):
    """ Return a new list containing all non-`None` arguments """
    return list(item for item in items if item is not None)


class SimpleNamespace(object):
    
    """ Implementation courtesy this SO answer:
        • https://stackoverflow.com/a/37161391/298171
    """
    __slots__ = tuplize('__dict__', '__weakref__')
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
    
    def __repr__(self):
        keys = sorted(self.__dict__)
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
        return "{}({})".format(type(self).__name__, ", ".join(items))
    
    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class Namespace(SimpleNamespace, collections.abc.MutableMapping,
                                 collections.abc.Hashable):
    
    """ A less-simple namespace -- one implementing several useful
        interface APIs from `collections.abc`.
    """
    __slots__ = tuple()
    
    def __bool__(self):
        return bool(self.__dict__)
    
    def __len__(self):
        return len(self.__dict__)
    
    def __contains__(self, key):
        return key in self.__dict__
    
    def __getitem__(self, key):
        return self.__dict__[key]
    
    def __setitem__(self, key, val):
        self.__dict__[key] = val
    
    def __delitem__(self, key):
        del self.__dict__[key]
    
    def get(self, key, default_value=None):
        return self.__dict__.get(key, default_value)
    
    def pop(self, key, default_value=None):
        return self.__dict__.pop(key, default_value)
    
    def update(self, iterable=None, **kwargs):
        self.__dict__.update(iterable or tuple(), **kwargs)
    
    def __hash__(self):
        return hash(tuple(self.__dict__.keys()) +
                    tuple(self.__dict__.values()))
    
    def __iter__(self):
        return iter(self.__dict__)
    
    def __copy__(self):
        return type(self)(**self.__dict__)
    
    def __eq__(self, other):
        return self.__dict__ == getattr(other, '__dict__', {})

def wrap_value(value):
    return lambda *args, **kwargs: value

none_function = wrap_value(None)

string_types = uniquify(type(''),
                        type(u''),
                        type(r''),
                       *six.string_types)

byte_types = tuplize({ bytes, bytearray } - { str })

class Memoizer(dict):
    
    """ Very simple memoizer (only works with positional args) """
    
    def __init__(self, function):
        super(Memoizer, self).__init__()
        self.original = function
    
    def __missing__(self, key):
        function = self.original
        self[key] = out = function(*key)
        return out
    
    @property
    def original(self):
        return self.original_function
    
    @original.setter
    def original(self, value):
        if not value:
            self.original_function = none_function
        elif not callable(value):
            self.original_function = wrap_value(value)
        else:
            self.original_function = value
    
    @property
    def __wrapped__(self):
        return self.original_function
    
    def __call__(self, function=None):
        if function is None:
            function = self.original
        else:
            self.original = function
        @wraps(function)
        def memoized(*args):
            return self[tuplize(*args)]
        memoized.__wrapped__ = function
        memoized.__instance__ = self
        return memoized


def memoize(function):
    memoinstance = Memoizer(function)
    @wraps(function)
    def memoized(*args):
        return memoinstance[tuplize(*args)]
    memoized.__wrapped__ = function
    memoized.__instance__ = memoinstance
    return memoized

def stringify(instance, fields):
    """ Stringify an object instance, using an iterable field list to
        extract and render its values, and printing them along with the 
        typename of the instance and its memory address -- yielding a
        repr-style string of the format:
        
            TypeName(fieldname="val", otherfieldname="otherval") @ 0x0FE
        
        The `stringify(…)` function is of use in `__str__()` and `__repr__()`
        definitions, E.G. something like:
        
            def __repr__(self):
                return stringify(self, type(self).__slots__)
        
    """
    field_dict = {}
    for field in fields:
        field_value = getattr(instance, field, "")
        field_value = callable(field_value) and field_value() or field_value
        if field_value:
            field_dict.update({ u8str(field) : field_value })
    field_dict_items = []
    for k, v in field_dict.items():
        field_dict_items.append('''%s="%s"''' % (k, v))
    typename = type(instance).__name__
    field_dict_string = ", ".join(field_dict_items)
    hex_id = hex(id(instance))
    return "%s(%s) @ %s" % (typename, field_dict_string, hex_id)

def suffix_searcher(suffix):
    """ Return a boolean function that will search for the given
        file suffix in strings with which it is called, returning
        True when they are found and False when they aren’t.
        
        Useful in filter(…) calls and comprehensions, e.g.:
        
        >>> plists = filter(suffix_searcher('plist'), os.listdir())
        >>> mmsuffix = suffix_searcher('mm')
        >>> objcpp = (f for f in os.listdir() where mmsuffix(f))
    """
    import re, os
    if len(suffix) < 1:
        return lambda searching_for: True
    regex_str = r""
    if suffix.startswith(os.extsep):
        regex_str += r"\%s$" % suffix
    else:
        regex_str += r"\%s%s$" % (os.extsep, suffix)
    searcher = re.compile(regex_str, re.IGNORECASE).search
    return lambda searching_for: bool(searcher(searching_for))

def u8encode(source):
    """ Encode a source as bytes using the UTF-8 codec """
    return bytes(source, encoding=UTF8_ENCODING)

def u8bytes(source):
    """ Encode a source as bytes using the UTF-8 codec, guaranteeing
        a proper return value without raising an error
    """
    if type(source) is bytes:
        return source
    elif type(source) is bytearray:
        return bytes(source)
    elif isinstance(source, string_types):
        return u8encode(source)
    elif isinstance(source, (int, float)):
        return u8encode(str(source))
    elif type(source) is bool:
        return source and b'True' or b'False'
    elif source is None:
        return b'None'
    return bytes(source)

def u8str(source):
    """ Encode a source as a Python string, guaranteeing a proper return
        value without raising an error
    """
    return type(source) is str and source \
                        or u8bytes(source).decode(UTF8_ENCODING)



#cython: cdivision=True
#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False

from __future__ import division

ctypedef unsigned char byte_t

cdef extern from "halftone.h" nogil:
    byte_t atkinson_add_error(int b, int e)

import numpy
cimport numpy
cimport cython

from instakit.utils import ndarrays

ctypedef numpy.int_t int_t
ctypedef numpy.uint8_t uint8_t
ctypedef numpy.uint32_t uint32_t
ctypedef numpy.float32_t float32_t

cdef void atkinson_dither(uint8_t[:, :] input_view,
                          int_t width, int_t height,
                          byte_t* threshold_matrix_ptr) nogil:
    
    cdef int_t y, x, err
    cdef uint8_t oldpx, newpx
    
    for y in range(height):
        for x in range(width):
            oldpx = input_view[y, x]
            newpx = <uint8_t>threshold_matrix_ptr[oldpx]
            err = (<int_t>oldpx - <int_t>newpx) >> 3
            
            input_view[y, x] = newpx
            
            if y + 1 < height:
                input_view[y+1, x] = atkinson_add_error(input_view[y+1, x], err)
            
            if y + 2 < height:
                input_view[y+2, x] = atkinson_add_error(input_view[y+2, x], err)
            
            if (y > 0) and (x + 1 < width):
                input_view[y-1, x+1] = atkinson_add_error(input_view[y-1, x+1], err)
            
            if x + 1 < width:
                input_view[y, x+1] = atkinson_add_error(input_view[y, x+1], err)
            
            if (y + 1 < height) and (x + 1 < width):
                input_view[y+1, x+1] = atkinson_add_error(input_view[y+1, x+1], err)
            
            if x + 2 < width:
                input_view[y, x+2] = atkinson_add_error(input_view[y, x+2], err)

cdef inline uint8_t floyd_steinberg_add_error_SEVEN(uint8_t base,
                                                      int_t err) nogil:
    cdef int_t something = <int_t>base + err * 7 / 16
    return <uint8_t>max(min(255, something), 0)

cdef inline uint8_t floyd_steinberg_add_error_THREE(uint8_t base,
                                                      int_t err) nogil:
    cdef int_t something = <int_t>base + err * 3 / 16
    return <uint8_t>max(min(255, something), 0)

cdef inline uint8_t floyd_steinberg_add_error_CINCO(uint8_t base,
                                                      int_t err) nogil:
    cdef int_t something = <int_t>base + err * 5 / 16
    return <uint8_t>max(min(255, something), 0)

cdef inline uint8_t floyd_steinberg_add_error_ALONE(uint8_t base,
                                                      int_t err) nogil:
    cdef int_t something = <int_t>base + err * 1 / 16
    return <uint8_t>max(min(255, something), 0)

cdef void floyd_steinberg_dither(uint8_t[:, :] input_view,
                                 int_t width, int_t height,
                                 byte_t* threshold_matrix_ptr) nogil:
    
    cdef int_t y, x, err
    cdef uint8_t oldpx, newpx
    
    for y in range(height):
        for x in range(width):
            oldpx = input_view[y, x]
            newpx = <uint8_t>threshold_matrix_ptr[oldpx]
            input_view[y, x] = newpx
            err = <int_t>oldpx - <int_t>newpx
            
            if (x + 1 < width):
                input_view[y, x+1] = floyd_steinberg_add_error_SEVEN(input_view[y, x+1], err)
            
            if (y + 1 < height) and (x > 0):
                input_view[y+1, x-1] = floyd_steinberg_add_error_THREE(input_view[y+1, x-1], err)
            
            if (y + 1 < height):
                input_view[y+1, x] = floyd_steinberg_add_error_CINCO(input_view[y+1, x], err)
            
            if (y + 1 < height) and (x + 1 < width):
                input_view[y+1, x+1] = floyd_steinberg_add_error_ALONE(input_view[y+1, x+1], err)

@cython.freelist(16)
cdef class ThresholdMatrixDitherer:
    
    """ Base ditherer image processor class """
    
    cdef:
        byte_t threshold_matrix[256]
    
    def __cinit__(self, float32_t threshold = 128.0):
        cdef uint8_t idx
        with nogil:
            for idx in range(255):
                self.threshold_matrix[idx] = <unsigned char>(<uint8_t>(<float32_t>idx / threshold) * 255)

cdef class Atkinson(ThresholdMatrixDitherer):
    
    """ Fast cythonized Atkinson-dither halftone image processor """
    
    def process(self, image not None):
        input_array = ndarrays.fromimage(image.convert('L'), dtype=numpy.uint8)
        cdef uint32_t width = image.size[0]
        cdef uint32_t height = image.size[1]
        cdef uint8_t[:, :] input_view = input_array
        with nogil:
            atkinson_dither(input_view, width, height, self.threshold_matrix)
        output_array = numpy.asarray(input_view.base)
        return ndarrays.toimage(output_array)

cdef class FloydSteinberg(ThresholdMatrixDitherer):
    
    """ Fast cythonized Floyd-Steinberg-dither halftone image processor """
    
    def process(self, image not None):
        input_array = ndarrays.fromimage(image.convert('L'), dtype=numpy.uint8)
        cdef uint32_t width = image.size[0]
        cdef uint32_t height = image.size[1]
        cdef uint8_t[:, :] input_view = input_array
        with nogil:
            floyd_steinberg_dither(input_view, width, height, self.threshold_matrix)
        output_array = numpy.asarray(input_view.base)
        return ndarrays.toimage(output_array)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#    
#    INSTAKIT -- Instagrammy image-processors and tools, based on Pillow and SciPy
#    
#    Copyright © 2012-2025 Alexander Böhn
#    
#    Permission is hereby granted, free of charge, to any person obtaining a copy 
#    of this software and associated documentation files (the "Software"), to deal 
#    in the Software without restriction, including without limitation the rights 
#    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
#    copies of the Software, and to permit persons to whom the Software is 
#    furnished to do so, subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be included in all 
#    copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
#    SOFTWARE.
#
from __future__ import print_function
from pkgutil import extend_path
from os.path import dirname

from clu.version import read_version_file, VersionInfo

# module exports:
__all__ = ('__version__', 'version_info',
           '__title__', '__author__', '__maintainer__',
           '__license__', '__copyright__')

__dir__ = lambda: list(__all__)

if '__path__' in locals():
    __path__ = extend_path(__path__, __name__)
    __all__ += ('__path__',)

# Embedded project metadata:
__version__ = read_version_file(dirname(__file__))
__title__ = 'instakit'
__author__ = 'Alexander Böhn'
__maintainer__ = __author__
__license__ = 'MIT'
__copyright__ = '© 2012-2025 %s' % __author__

# The CLU project version:
version_info = VersionInfo(__version__)



#!/usr/bin/env python
# encoding: utf-8
"""
blur.py

Created by FI$H 2000 on 2012-08-23.
Copyright (c) 2012 Objects In Space And Time, LLC. All rights reserved.
"""
from __future__ import print_function

from PIL import ImageFilter
from instakit.abc import Processor

class ImagingCoreFilterMixin(Processor):
    """ A mixin furnishing a `process(…)` method to PIL.ImageFilter classes """
    __slots__ = tuple() # a __dict__ slot exists already
    
    def process(self, image):
        return image.filter(self)

class Contour(ImageFilter.CONTOUR, ImagingCoreFilterMixin):
    """ Contour-Enhance Filter """
    __slots__ = tuple()


class Detail(ImageFilter.DETAIL, ImagingCoreFilterMixin):
    """ Detail-Enhance Filter """
    __slots__ = tuple()


class Emboss(ImageFilter.EMBOSS, ImagingCoreFilterMixin):
    """ Emboss-Effect Filter """
    __slots__ = tuple()


class FindEdges(ImageFilter.FIND_EDGES, ImagingCoreFilterMixin):
    """ Edge-Finder Filter """
    __slots__ = tuple()


class EdgeEnhance(ImageFilter.EDGE_ENHANCE, ImagingCoreFilterMixin):
    """ Edge-Enhance Filter """
    __slots__ = tuple()


class EdgeEnhanceMore(ImageFilter.EDGE_ENHANCE_MORE, ImagingCoreFilterMixin):
    """ Edge-Enhance (With Extreme Predjudice) Filter """
    __slots__ = tuple()


class Smooth(ImageFilter.SMOOTH, ImagingCoreFilterMixin):
    """ Image-Smoothing Filter """
    __slots__ = tuple()


class SmoothMore(ImageFilter.SMOOTH_MORE, ImagingCoreFilterMixin):
    """ Image-Smoothing (With Extreme Prejudice) Filter """
    __slots__ = tuple()


class Sharpen(ImageFilter.SHARPEN, ImagingCoreFilterMixin):
    """ Image Sharpener """
    __slots__ = tuple()


class UnsharpMask(ImageFilter.UnsharpMask, ImagingCoreFilterMixin):
    """ Unsharp Mask Filter 
        Optionally initialize with params:
            radius (2), percent (150), threshold (3) """
    __slots__ = tuple()


class SimpleGaussianBlur(ImageFilter.GaussianBlur, ImagingCoreFilterMixin):
    """ Simple Gaussian Blur Filter 
        Optionally initialize with radius (2) """
    __slots__ = tuple()


class GaussianBlur(Processor):
    """ Gaussian Blur Filter 
        Optionally initialize with params:
            sigmaX (3)
            sigmaY (3; same as sigmaX)
            sigmaZ (0; same as sigmaX)
    """
    __slots__ = ('sigmaX', 'sigmaY', 'sigmaZ')
    
    def __init__(self, sigmaX=3, sigmaY=None, sigmaZ=None):
        self.sigmaX = sigmaX
        self.sigmaY = sigmaY or sigmaX
        self.sigmaZ = sigmaZ or sigmaX
    
    def process(self, image):
        from PIL import Image
        from numpy import array
        from instakit.utils import kernels
        return Image.fromarray(kernels.gaussian_blur_filter(
                                input=array(image),
                               sigmaX=self.sigmaX,
                               sigmaY=self.sigmaY,
                               sigmaZ=self.sigmaZ))


if __name__ == '__main__':
    from instakit.utils.static import asset
    from instakit.utils.mode import Mode
    
    image_paths = list(map(
        lambda image_file: asset.path('img', image_file),
            asset.listfiles('img')))
    image_inputs = list(map(
        lambda image_path: Mode.RGB.open(image_path),
            image_paths))
    
    for image_input in image_inputs:
        # image_input.show()
        # Contour().process(image_input).show()
        # Detail().process(image_input).show()
        # Emboss().process(image_input).show()
        # EdgeEnhance().process(image_input).show()
        # EdgeEnhanceMore().process(image_input).show()
        # FindEdges().process(image_input).show()
        # Smooth().process(image_input).show()
        # SmoothMore().process(image_input).show()
        # Sharpen().process(image_input).show()
        # UnsharpMask().process(image_input).show()
        GaussianBlur(sigmaX=3).process(image_input).show()
        # SimpleGaussianBlur(radius=3).process(image_input).show()
    
    print(image_paths)
    



#!/usr/bin/env python
# encoding: utf-8
"""
stats.py

Created by FI$H 2000 on 2018-12-24.
Copyright (c) 2018 Objects In Space And Time, LLC. All rights reserved.
"""
from __future__ import print_function
from instakit.utils.mode import Mode

def pixel_count(image):
    """ Return the number of pixels in the input image. """
    width, height = image.size
    return width * height

def color_count(image):
    """ Return the number of color values in the input image --
        this is the number of pixels times the band count
        of the image.
    """
    width, height = image.size
    return width * height * Mode.of(image).band_count

def histogram_sum(image):
    """ Return the sum of the input images’ histogram values --
        Basically this is an optimized way of doing:
            
            out = 0.0
            for value, count in enumerate(histogram):
                out += value * count
            return out
            
        … the one-liner uses the much faster sum(…) in léu
        of looping over the histogram’s enumerated values.
    """
    histogram = Mode.L.process(image).histogram()
    return sum(value * count for value, count in enumerate(histogram))

def histogram_mean(image):
    """ Return the mean of the input images’ histogram values. """
    return float(histogram_sum(image)) / pixel_count(image)

def histogram_entropy_py(image):
    """ Calculate the entropy of an images' histogram.
        Used for “smart cropping” in easy-thumbnails:
            https://git.io/fhqxd
    """
    from math import log2, fsum
    histosum = float(color_count(image))
    histonorm = (histocol / histosum for histocol in image.histogram())
    
    return -fsum(p * log2(p) for p in histonorm if p != 0.0)

from PIL import Image

histogram_entropy = hasattr(Image.Image, 'entropy') \
                        and Image.Image.entropy \
                        or histogram_entropy_py



#!/usr/bin/env python
# encoding: utf-8
"""
squarecrop.py

Created by FI$H 2000 on 2012-08-23.
Copyright (c) 2012 Objects In Space And Time, LLC. All rights reserved.
"""
from __future__ import print_function

from instakit.abc import Processor

class SquareCrop(Processor):
    
    """ Crop an image to an Instagrammy square, by whittling away
        the parts of the image with the least entropy.
        
        Based on a smart-crop implementation from easy-thumbnails:
            https://git.io/fhqxj
    """
    __slots__ = tuple()
    
    @staticmethod
    def compare_entropy(start_slice, end_slice, slice, difference):
        """ Calculate the entropy of two slices (from the start and end
            of an axis), returning a tuple containing the amount that
            should be added to the start, and removed from the end
            of that axis.
            
            Based on the eponymous function from easy-thumbnails:
                https://git.io/fhqpT
        """
        from instakit.utils.stats import histogram_entropy
        
        start_entropy = histogram_entropy(start_slice)
        end_entropy = histogram_entropy(end_slice)
        
        if end_entropy and abs(start_entropy / end_entropy - 1) < 0.01:
            # Less than 1% difference, remove from both sides.
            if difference >= slice * 2:
                return slice, slice
            half_slice = slice // 2
            return half_slice, slice - half_slice
        
        if start_entropy > end_entropy:
            return 0, slice
        else:
            return slice, 0
    
    def process(self, image):
        source_x, source_y = image.size
        target_width = target_height = min(image.size)
        
        diff_x = int(source_x - min(source_x, target_width))
        diff_y = int(source_y - min(source_y, target_height))
        left = top = 0
        right, bottom = source_x, source_y
        
        while diff_x:
            slice = min(diff_x, max(diff_x // 5, 10))
            start = image.crop((left, 0, left + slice, source_y))
            end = image.crop((right - slice, 0, right, source_y))
            add, remove = self.compare_entropy(start, end, slice, diff_x)
            left += add
            right -= remove
            diff_x = diff_x - add - remove
        
        while diff_y:
            slice = min(diff_y, max(diff_y // 5, 10))
            start = image.crop((0, top, source_x, top + slice))
            end = image.crop((0, bottom - slice, source_x, bottom))
            add, remove = self.compare_entropy(start, end, slice, diff_y)
            top += add
            bottom -= remove
            diff_y = diff_y - add - remove
        
        box = (left, top, right, bottom)
        return image.crop(box)


if __name__ == '__main__':
    from instakit.utils.static import asset
    from instakit.utils.mode import Mode
    
    image_paths = list(map(
        lambda image_file: asset.path('img', image_file),
            asset.listfiles('img')))
    image_inputs = list(map(
        lambda image_path: Mode.RGB.open(image_path),
            image_paths))
    
    for image_input in image_inputs:
        image_input.show()
        SquareCrop().process(image_input).show()
    
    print(image_paths)

